'''

test_disable_cache.py: Testing disabled cache function

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import os
import re
import sys
sys.path.append('..')  # noqa

from unittest import TestCase
import shutil
import tempfile

VERSION = sys.version_info[0]

print("*** PYTHON VERSION %s CLIENT TESTING START ***" % VERSION)


class TestDisabledCache(TestCase):

    def setUp(self):
        os.environ['SINGULARITY_DISABLE_CACHE'] = "yes"
        self.tmpdir = tempfile.mkdtemp()
        os.environ['SINGULARITY_ROOTFS'] = self.tmpdir
        print("\n---START----------------------------------------")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        print("---END------------------------------------------")

    def test_cache_disabled(self):
        print("Case 3: Testing that cache gets disabled.")
        from defaults import DISABLE_CACHE
        self.assertEqual(DISABLE_CACHE, True)


if __name__ == '__main__':
    unittest.main()



'''

main.py: Singularity Hub helper functions for Singularity in Python

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) # noqa

from shell import parse_image_uri

from .api import (
    extract_metadata,
    get_image_name,
    SingularityApiConnection
)

from sutils import (
    get_cache,
    write_file
)

from defaults import SHUB_PREFIX

from message import bot
import json
import re
import os


def SIZE(image, contentfile=None):
    '''size is intended to be run before an import,
    to return to the contentfile a list of sizes
    (one per layer) corresponding with the layers
    that will be downloaded for image
    '''
    bot.debug("Starting SHub SIZE, will get size from manifest")
    bot.debug("Singularity Hub image: %s" % image)
    client = SingularityApiConnection(image=image)
    manifest = client.get_manifest()
    if 'size_mb' in manifest:  # sregistry
        size = manifest['size_mb']
    else:
        size = manifest['metrics']['size']
    if contentfile is not None:
        write_file(contentfile, str(size), mode="w")
    return size


def PULL(image, download_folder=None, layerfile=None):
    '''PULL will retrieve a Singularity Hub image and
    download to the local file system, to the variable
    specified by SINGULARITY_PULLFOLDER.
    :param image: the singularity hub image name
    :param download folder: the folder to pull the image to.
    :param layerfile: if defined, write pulled image to file
    '''
    client = SingularityApiConnection(image=image)
    manifest = client.get_manifest()

    if download_folder is None:
        cache_base = get_cache(subfolder="shub")
    else:
        cache_base = os.path.abspath(download_folder)

    bot.debug("Pull folder set to %s" % cache_base)

    # The image name is the md5 hash, download if it's not there
    image_name = get_image_name(manifest)

    # Did the user specify an absolute path?
    custom_folder = os.path.dirname(image_name)
    if custom_folder not in [None, ""]:
        cache_base = custom_folder
        image_name = os.path.basename(image_name)

    image_file = "%s/%s" % (cache_base, image_name)

    bot.debug('Pulling to %s' % image_file)
    if not os.path.exists(image_file):
        image_file = client.download_image(manifest=manifest,
                                           download_folder=cache_base)
    else:
        if not bot.is_quiet():  # not --quiet
            print("Image already exists at %s, skipping download" % image_file)

    if not bot.is_quiet():  # not --quiet
        print("Singularity Hub Image Download: %s" % image_file)

    manifest = {'image_file': image_file,
                'manifest': manifest,
                'cache_base': cache_base,
                'image': image}

    if layerfile is not None:
        bot.debug("Writing Singularity Hub image path to %s" % layerfile)
        write_file(layerfile, image_file, mode="w")

    return manifest


def IMPORT(image, layerfile, labelfile=None):
    '''IMPORT takes one more step than ADD, returning
    the image written to a layerfile plus metadata written
    to the metadata base in rootfs.
    :param image: the singularity hub image name
    '''
    manifest = PULL(image, layerfile=layerfile)

    # Write metadata to base
    manifest['metadata'] = extract_metadata(manifest=manifest['manifest'],
                                            labelfile=labelfile,
                                            prefix="%s_" % SHUB_PREFIX)
    return manifest



'''

main.py: main utility functions for json module in Singularity python API

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),
                os.path.pardir)))  # noqa

from sutils import (
    print_json,
    read_file,
    read_json,
    write_json
)

from defaults import (
    ENVIRONMENT,
    HELPFILE,
    LABELFILE,
    RUNSCRIPT,
    TESTFILE,
    DEFFILE
)

from message import bot
import json
import re


def DUMP(jsonfile):
    '''DUMP will return the entire layfile as text, key value pairs
    :param jsonfile_path: the path to the jsonfile
    '''
    bot.debug("Reading %s to prepare dump to STDOUT" % jsonfile)
    if not os.path.exists(jsonfile):
        bot.error("Cannot find %s, exiting." % jsonfile)
        sys.exit(1)

    contents = read_json(jsonfile)
    dump = ""
    for key, value in contents.items():
        dump = '%s%s:"%s"\n' % (dump, key, value)
    dump = dump.strip('\n')
    print(dump)
    return dump


def GET(key, jsonfile):
    '''GET will return a key from the jsonfile
    if it exists. If it doesn't, returns None.
    '''
    key = format_keyname(key)
    bot.debug("GET %s from %s" % (key, jsonfile))
    if not os.path.exists(jsonfile):
        bot.error("Cannot find %s, exiting." % jsonfile)
        sys.exit(1)

    contents = read_json(jsonfile)
    if key in contents:
        value = contents[key]
        print(value)
        bot.debug('%s is %s' % (key, value))
    else:
        bot.error("%s is not defined in file. Exiting" % key)
        sys.exit(1)
    return value


def ADD(key, value, jsonfile, force=False, quiet=False):
    '''ADD will write or update a key in a json file
    '''

    # Check that key is not empty
    if key.strip() in ['#', '', None]:
        bot.verbose('Empty key %s, skipping' % key)
        sys.exit(0)

    key = format_keyname(key)
    bot.debug("Adding label: '%s' = '%s'" % (key, value))
    bot.debug("ADD %s from %s" % (key, jsonfile))
    if os.path.exists(jsonfile):
        contents = read_json(jsonfile)
        if key in contents:
            msg = 'Warning, %s is already set. ' % key
            msg += 'Overwrite is set to %s' % force
            if not quiet:
                bot.debug(msg)
            if force is True:
                contents[key] = value
            else:
                msg = '%s found in %s ' % (key, jsonfile)
                msg += 'and overwrite set to %s.' % force
                bot.error(msg)
                sys.exit(1)
        else:
            contents[key] = value
    else:
        contents = {key: value}
    bot.debug('%s is %s' % (key, value))
    write_json(contents, jsonfile)
    return value


def DELETE(key, jsonfile):
    '''DELETE will remove a key from a json file
    '''
    key = format_keyname(key)
    bot.debug("DELETE %s from %s" % (key, jsonfile))
    if not os.path.exists(jsonfile):
        bot.error("Cannot find %s, exiting." % jsonfile)
        sys.exit(1)

    contents = read_json(jsonfile)
    if key in contents:
        del contents[key]
        if len(contents) > 0:
            write_json(contents, jsonfile)
        else:
            bot.debug('%s is empty, deleting.' % jsonfile)
            os.remove(jsonfile)
        return True
    else:
        bot.debug('Warning, %s not found in %s' % (key, jsonfile))
        return False


def INSPECT(inspect_labels=None,
            inspect_def=None,
            inspect_app=None,
            inspect_runscript=None,
            inspect_test=None,
            inspect_help=None,
            inspect_env=None,
            pretty_print=True):
    '''INSPECT will print a "manifest" for an image, with one or more
    variables asked for by the user. The default prints a human readable
    format (text and json) and if pretty_print  is turned to False,
    the entire thing will be returned as json via the JSON API standard.
    The base is checked for the metadata folder, and if it does not exist,
    the links are searched for and parsed (to support older versions).
    For all of the following, each is returned if it exists

    :param inspect_runscript: if not None, will include runscript
    :param inspect_labels: if not None, will include labels
    :param inspect_def: if not None, will include definition file
    :param inspect_test: if not None, will include test
    :param inspect_env: if not None, will include environment
    :param inspect_help: if not None, include helpfile
    :param pretty_print: if False, return all JSON API spec
    '''
    from defaults import (LABELFILE, HELPFILE, RUNSCRIPT,
                          TESTFILE, ENVIRONMENT)

    if inspect_app is not None:
        LABELBASE = "scif/apps/%s/scif/labels.json" % inspect_app
        LABELFILE = LABELFILE.replace('.singularity.d/labels.json', LABELBASE)
        HELPBASE = "scif/apps/%s/scif/runscript.help" % inspect_app
        HELPFILE = HELPFILE.replace('.singularity.d/runscript.help', HELPBASE)
        RUNBASE = "scif/apps/%s/scif/runscript" % inspect_app
        RUNSCRIPT = "%s/%s" % (RUNSCRIPT.strip('singularity'), RUNBASE)

        TESTBASE = "scif/apps/%s/scif/test" % inspect_app
        TESTFILE = TESTFILE.replace(".singularity.d/test", TESTBASE)
        ENVBASE = "scif/apps/%s/scif/" % inspect_app
        ENVIRONMENT = ENVIRONMENT.replace(".singularity.d/", ENVBASE)

    data = dict()
    errors = dict()

    # Labels
    if inspect_labels:
        bot.verbose2("Inspection of labels selected.")
        if os.path.exists(LABELFILE):
            data["labels"] = read_json(LABELFILE)
        else:
            data["labels"] = None
            error_detail = "This container does not have labels"
            errors["labels"] = generate_error(404,
                                              detail=error_detail,
                                              title="Labels Undefined")

    # Helpfile
    if inspect_help:
        bot.verbose2("Inspection of helpfile selected.")
        if os.path.exists(HELPFILE):
            data["help"] = read_file(HELPFILE, readlines=False)
        else:
            data["help"] = None
            error_detail = "This container does not have a helpfile"
            errors["help"] = generate_error(404,
                                            detail=error_detail,
                                            title="Help Undefined")

    # Definition File
    if inspect_def:
        bot.verbose2("Inspection of deffile selected.")
        if os.path.exists(DEFFILE):
            data["deffile"] = read_file(DEFFILE, readlines=False)
        else:
            error_detail = "This container doesn't have a bootstrap recipe"
            data["deffile"] = None
            errors["deffile"] = generate_error(404,
                                               title="Recipe Undefined",
                                               detail=error_detail)

    # Runscript
    if inspect_runscript:
        bot.verbose2("Inspection of runscript selected.")
        if os.path.exists(RUNSCRIPT):
            data["runscript"] = read_file(RUNSCRIPT, readlines=False)
        else:
            error_detail = "This container does not have any runscript defined"
            data["runscript"] = None
            errors["runscript"] = generate_error(404,
                                                 title="Runscript Undefined",
                                                 detail=error_detail)

    # Test
    if inspect_test:
        bot.verbose2("Inspection of test selected.")
        if os.path.exists(TESTFILE):
            data["test"] = read_file(TESTFILE, readlines=False)
        else:
            error_detail = "This container does not have any tests defined"
            data["test"] = None
            errors["test"] = generate_error(404,
                                            title="Tests Undefined",
                                            detail=error_detail)

    # Environment
    if inspect_env:
        bot.verbose2("Inspection of environment selected.")
        if os.path.exists(ENVIRONMENT):
            data["environment"] = read_file(ENVIRONMENT, readlines=False)
        else:
            error_detail = "This container doesn't have environment defined"
            data["environment"] = None
            errors["environment"] = generate_error(404,
                                                   title="Tests Undefined",
                                                   detail=error_detail)

    if pretty_print:
        bot.verbose2("Structured printing specified.")
        for dtype, content in data.items():
            if content is not None:
                if isinstance(content, dict):
                    print_json(content, print_console=True)
                else:
                    bot.info(content)
            else:
                print_json(errors[dtype], print_console=True)

    else:
        bot.verbose2("Unstructed printing specified")
        # Only define a type if there is data to return
        # else return errors
        result = dict()
        if len(data) > 0:
            result["data"] = {"attributes": data,
                              "type": "container"}
        else:
            result["errors"] = []
            for dtype, content in errors.items():
                result["errors"].append(content)
        print_json(result, print_console=True)


def generate_error(status=None, title=None, detail=None):
    '''generate_error will return a JSON API error object,
    intended to be added to the list of "errors" at the top level.
    Errors and data cannot both be returned, and so these data
     structures are only returned given no data.
    :param status: the status code to return (http status codes)
    :param title: a short title to describe the error
    :param detail: detail about the error
    '''
    params = locals()
    error = dict()
    for key, value in params.items():
        if value is not None:
            error[key] = value
    return error


def format_keyname(key):
    '''format keyname will ensure that all keys
    are uppcase, with no special characters
    '''
    if key.startswith('org.label-schema'):
        return re.sub('[^A-Za-z0-9-.]+', '_', key).lower()
    return re.sub('[^A-Za-z0-9]+', '_', key).upper()



'''

test_base.py: Test client initialized with python modules

Copyright (c) 2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import os
import re
import sys
import tarfile
sys.path.append('..')  # noqa

from unittest import TestCase
from base import ApiConnection
import shutil
import tempfile

VERSION = sys.version_info[0]

print("*** PYTHON VERSION %s BASE TESTING START ***" % VERSION)


class TestShell(TestCase):

    def setUp(self):

        self.client = ApiConnection()
        print("\n---START----------------------------------------")

    def tearDown(self):
        print("---END------------------------------------------")

    def test_client_headers(self):
        '''test_load_client will load an empty client
        '''
        print("Testing client default headers")
        required = ['Accept', 'Content-Type']
        for required_header in required:
            self.assertTrue(required_header in self.client.headers)
        self.assertEqual(self.client.update_token(), None)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) 2017, SingularityWare, LLC. All rights reserved.
# Copyright (c) 2017, Vanessa Sochat. All rights reserved.
#
# See the COPYRIGHT.md file at the top-level directory of this
# distribution and at https://github.com/singularityware/singularity
#
# This file is part of the Singularity Linux container project.
# It is subject to the license terms in the LICENSE.md file
# found in the top-level directory of this distribution and at
# https://github.com/singularityware/singularity. No part
# of Singularity, including this file, may be copied, modified,
# propagated, or distributed except according to the terms
# contained in the LICENSE.md file.


import os
import platform
import sys
import re
try:
    from urllib.request import urlopen, Request, unquote
except ImportError:
    from urllib2 import urlopen, Request, HTTPError


os_base, os_name, os_version = platform.linux_distribution()
os_base = os_base.lower()
os_names = "|".join([x.lower() for x in os_name.split('/')])
base = os.environ["SINGULARITY_ROOTFS"]
os.chdir(base)

##################################################################
# Common Vulnerabilities Database, High Risk
##################################################################

base = "https://security-tracker.debian.org/tracker/status/release"
filters = "?filter=1&filter=high_urgency"
release = "stable"

url = Request('%s/%s/%s' % (base, release, filters))
response = urlopen(url).read().decode('utf-8')
cve_codes = re.findall(">CVE-(.*?)<", response)

returncode = 0

# We are only testing debian
if os_base not in ['debian', 'ubuntu']:
    print("OS not in debian/ubuntu family, skipping test.")
    sys.exit(returncode)

# Iterate through the CVE codes, and assess if the distribution matches
print("Checking %s system for %s CVE vulnerabilities..." % (os_base,
                                                            len(cve_codes)))
for cve_code in cve_codes:

    url = "https://security-tracker.debian.org/tracker/CVE-%s" % cve_code
    request = Request(url)
    try:
        response = urlopen(request)
    except HTTPError:
        pass

    html = response.read().decode('utf-8')
    table = html.replace('PTS', '').split('<table>')[2]
    title = table.split('<tr>')[2]
    title = re.findall('">(.*?)</a>', title)[0]

    print("CVE-%s: %s" % (cve_code, title))

    rows = table.replace('</td>', '').split('<tr>')
    for row in rows:
        if row:
            if re.search(os_names, row):
                print("PROBLEM:  Vulnerability CVE-%s" % cve_code)
                print("RESOLVE:  %s" % url)
                returncode = 1

sys.exit(returncode)



#!/usr/bin/env python

'''

dump.py: wrapper for dump a a json file for Singularity Hub command line tool.

This function takes input arguments of the following:

   --file: should be the json file to read

Copyright (c) 2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),
                os.path.pardir,
                os.path.pardir)))  # noqa

import optparse
from helpers.json.main import DUMP
from message import bot


def get_parser():

    parser = optparse.OptionParser(description="DUMP json")

    parser.add_option("--file",
                      dest='file',
                      help="Path to json file to dump from",
                      type=str,
                      default=None)

    return parser


def main():

    parser = get_parser()

    try:
        (args, options) = parser.parse_args()
    except Exception:
        sys.exit(0)

    if args.file is not None:

        dump = DUMP(args.file)

    else:
        bot.error("--file must be defined for DUMP. Exiting")
        sys.exit(1)


if __name__ == '__main__':
    main()



'''

python: base template for making a connection to an API

Copyright (c) 2017, Vanessa Sochat. All rights reserved.

'''

from message import bot
from defaults import SINGULARITY_WORKERS
import multiprocessing
import itertools
import tempfile
import time
import signal
import sys
import re
import os


try:
    from urllib.parse import urlencode, urlparse
    from urllib.request import urlopen, Request, unquote
    from urllib.error import HTTPError
except ImportError:
    from urllib import urlencode, unquote
    from urlparse import urlparse
    from urllib2 import urlopen, Request, HTTPError


class MultiProcess(object):

    def __init__(self, workers=None):

        if workers is None:
            workers = SINGULARITY_WORKERS
        self.workers = workers
        bot.debug("Using %s workers for multiprocess." % (self.workers))

    def start(self):
        bot.debug("Starting multiprocess")
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()
        self.runtime = self.runtime = self.end_time - self.start_time
        bot.debug("Ending multiprocess, runtime: %s sec" % (self.runtime))

    def run(self, func, tasks, func2=None):
        '''run will send a list of tasks,
        a tuple with arguments, through a function.
        the arguments should be ordered correctly.
        :param func: the function to run with multiprocessing.pool
        :param tasks: a list of tasks, each a tuple
                      of arguments to process
        :param func2: filter function to run result
                      from func through (optional)
        '''

        # Keep track of some progress for the user
        progress = 1
        total = len(tasks)

        # If two functions are run per task, double total jobs
        if func2 is not None:
            total = total * 2

        finished = []
        level1 = []
        results = []

        try:
            prefix = "[%s/%s]" % (progress, total)
            bot.show_progress(0, total, length=35, prefix=prefix)
            pool = multiprocessing.Pool(self.workers, init_worker)

            self.start()
            for task in tasks:
                result = pool.apply_async(multi_wrapper,
                                          multi_package(func, [task]))
                results.append(result)
                level1.append(result._job)

            while len(results) > 0:
                result = results.pop()
                result.wait()
                bot.show_progress(progress, total, length=35, prefix=prefix)
                progress += 1
                prefix = "[%s/%s]" % (progress, total)

                # Pass the result through a second function?
                if func2 is not None and result._job in level1:
                    result = pool.apply_async(multi_wrapper,
                                              multi_package(func2,
                                                            [(result.get(),)]))
                    results.append(result)
                else:
                    finished.append(result.get())

            self.end()
            pool.close()
            pool.join()

        except (KeyboardInterrupt, SystemExit):
            bot.error("Keyboard interrupt detected, terminating workers!")
            pool.terminate()
            sys.exit(1)

        except Exception as e:
            bot.error(e)

        return finished


# Supporting functions for MultiProcess
def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def multi_wrapper(func_args):
    function, args = func_args
    return function(*args)


def multi_package(func, args):
    return zip(itertools.repeat(func), args)


class ApiConnection(object):

    def __init__(self, **kwargs):
        self.headers = dict()
        self.update_headers()

    def get_headers(self):
        return self.headers

    def _init_headers(self):
        return {'Accept': 'application/json',
                'Content-Type': 'application/json; charset=utf-8'}

    def update_headers(self, fields=None):
        '''get_headers will return a simple default
           header for a json post.
           This function will be adopted as needed.
        '''
        if len(self.headers) == 0:
            headers = self._init_headers()
        else:
            headers = self.headers

        if fields is not None:
            for key, value in fields.items():
                headers[key] = value

        header_names = ",".join(list(headers.keys()))
        bot.debug("Headers found: %s" % (header_names))
        self.headers = headers

    def update_token(self, response=None):
        '''empty update_token to be defined
           by subclasses, if necessary
        '''
        return None

    def stream(self,
               url,
               file_name,
               data=None,
               headers=None,
               default_headers=True,
               show_progress=False):

        '''stream is a get that will stream to file_name
        :param data: a dictionary of key:value items
                     to add to the data args variable
        :param url: the url to get
        :param show_progress: if True, show a progress bar with the bot
        :returns response: the requests response object, or stream
        '''
        bot.debug("GET (stream) %s" % (url))

        # If we use default headers, start with client's
        request_headers = dict()
        if default_headers and len(self.headers) > 0:
            request_headers = self.headers

        if headers is not None:
            request_headers.update(headers)

        request = self.prepare_request(headers=request_headers,
                                       data=data,
                                       url=url)

        response = self.submit_request(request)

        # Keep user updated with Progress Bar?
        if show_progress:
            content_size = None
            not_error = response.code not in [400, 401]
            if 'Content-Length' in response.headers and not_error:
                progress = 0
                content_size = int(response.headers['Content-Length'])
                bot.show_progress(progress, content_size, length=35)

        chunk_size = 1 << 20
        with open(file_name, 'wb') as filey:
            while True:
                chunk = response.read(chunk_size)
                if not chunk:
                    break
                try:
                    filey.write(chunk)
                    if show_progress:
                        if content_size is not None:
                            progress += chunk_size
                            bot.show_progress(iteration=progress,
                                              total=content_size,
                                              length=35,
                                              carriage_return=False)
                except Exception as error:
                    msg = "Error writing to %s:" % (file_name)
                    msg += " %s exiting" % (error)
                    bot.error(msg)
                    sys.exit(1)

            # Newline to finish download
            if show_progress:
                sys.stdout.write('\n')

        return file_name

    def get(self,
            url,
            data=None,
            headers=None,
            default_headers=True,
            return_response=False):

        '''get will use requests to get a particular url
        :param data: a dictionary of key:value
                     items to add to the data args variable
        :param url: the url to get
        :returns response: the requests response object, or stream
        '''

        bot.debug("GET %s" % (url))

        # If we use default headers, start with client's
        request_headers = dict()
        if default_headers and len(self.headers) > 0:
            request_headers = self.headers

        if headers is not None:
            request_headers.update(headers)

        request = self.prepare_request(headers=request_headers,
                                       data=data,
                                       url=url)

        response = self.submit_request(request,
                                       return_response=return_response)

        if return_response is True:
            return response

        return response.read().decode('utf-8')

    def submit_request(self, request, return_response=False):
        '''submit_request will make the request,
        via a stream or not. If return response is True, the
        response is returned as is without further parsing.
        Given a 401 error, the update_token function is called
        to try the request again, and only then the error returned.
        '''

        try:
            response = urlopen(request)

        # If we have an HTTPError, try to follow the response
        except HTTPError as error:

            # Case 1: we have an http 401 error, and need to refresh token
            bot.debug('Http Error with code %s' % (error.code))

            if error.code == 401:
                self.update_token(response=error)
                try:
                    request = self.prepare_request(request.get_full_url(),
                                                   headers=self.headers)
                    response = urlopen(request)
                except HTTPError as error:
                    bot.debug('Http Error with code %s' % (error.code))
                    return error
            else:
                return error

        return response

    def prepare_request(self, url, data=None, headers=None):
        '''prepare the request object, determining
        if there is data (making it
        a POST) or if we should stream the result.
        '''
        if data is not None:
            args = urlencode(data)

            request = Request(url=url,
                              data=args,
                              headers=headers)
        else:
            request = Request(url=url,
                              headers=headers)
        return request

    def download_atomically(self,
                            url,
                            file_name,
                            headers=None,
                            show_progress=False):

        '''download stream atomically will stream to a temporary file, and
        rename only upon successful completion. This is to ensure that
        errored downloads are not found as complete in the cache
        :param file_name: the file name to stream to
        :param url: the url to stream from
        :param headers: additional headers to add to the get (default None)
        '''
        try:
            tmp_file = "%s.%s" % (file_name,
                                  next(tempfile._get_candidate_names()))
            response = self.stream(url,
                                   file_name=tmp_file,
                                   headers=headers,
                                   show_progress=show_progress)
            os.rename(tmp_file, file_name)

        except Exception:

            download_folder = os.path.dirname(os.path.abspath(file_name))
            msg = "Error downloading %s. " % (url)
            msg += "Do you have permission to write to %s?" % (download_folder)
            bot.error(msg)
            try:
                os.remove(tmp_file)
            except Exception:
                pass
            sys.exit(1)

        return file_name



#!/usr/bin/env python

'''

size.py: python helper for Singularity size

ENVIRONMENTAL VARIABLES that are required for this executable:

    SINGULARITY_CONTAINER
    SINGULARITY_CONTENTS

For Docker, layer sizes are determined from the tarballs, and written
to the SINGULARITY_CONTENTS (contentfile). For Singularity Hub,
the image size is read from the manifest

Copyright (c) 2017, Vanessa Sochat. All rights reserved. 

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.
 
This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.
 
NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so. 


'''

import sys

from shell import (
    get_image_uri,
    remove_image_uri
)

from defaults import getenv
from message import bot


def main():
    '''this function will run the main size functions and call shub clients 
    '''

    container = getenv("SINGULARITY_CONTAINER",required=True)
    image_uri = get_image_uri(container)    
    container = remove_image_uri(container)
    from defaults import LAYERFILE

    ##############################################################################
    # Singularity Hub ############################################################
    ##############################################################################

    if image_uri == "shub://":

        bot.debug("\n*** STARTING SINGULARITY HUB SIZE PYTHON  ****")    
        from shub.main import SIZE
        SIZE(image=container,
             contentfile=LAYERFILE)

    elif image_uri == "docker://":

        from sutils import basic_auth_header
        from docker.main import SIZE

        bot.debug("Docker sizes will be written to: %s" %LAYERFILE)
        username = getenv("SINGULARITY_DOCKER_USERNAME") 
        password = getenv("SINGULARITY_DOCKER_PASSWORD",silent=True)

        auth = None
        if username is not None and password is not None:
            auth = basic_auth_header(username, password)

        SIZE(image=container,
             auth=auth,
             contentfile=LAYERFILE)

    else:
        bot.error("uri %s is not a currently supported uri for size. Exiting." %image_uri)
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Copyright (c) 2017, SingularityWare, LLC. All rights reserved.
# Copyright (c) 2017, Vanessa Sochat. All rights reserved.
#
# See the COPYRIGHT.md file at the top-level directory of this
# distribution and at https://github.com/singularityware/singularity
#
# This file is part of the Singularity Linux container project.
# It is subject to the license terms in the LICENSE.md file
# found in the top-level directory of this distribution and at
# https://github.com/singularityware/singularity. No part
# of Singularity, including this file, may be copied, modified,
# propagated, or distributed except according to the terms
# contained in the LICENSE.md file.
#
# Alert the user about files in the cache

import platform
import sys
import os

base = os.environ["SINGULARITY_ROOTFS"]
os.chdir(base)

os_base, os_name, os_version = platform.linux_distribution()
os_base = os_base.lower()

returncode = 0


def check_cache(returncode):
    '''check cache will look for archives in /var/cache, and
    return 1 if files are found'''

    # The cache should only have apt debconf ldconfig
    skip = ["apt", "debconf", "ldconfig"]
    cache_dirs = [x for x in os.listdir("var/cache")
                  if x not in skip]
    if len(cache_dirs) > 3:
        to_remove = "\n".join(["rm -rf /var/cache/%s" % x for x in cache_dirs])
        print("PROBLEM:  /var/cache has uneccessary entries")
        print("RESOLVE:  \n%s" % to_remove)
        returncode = 1


def check_apt(returncode):
    '''check apt will look for files in apt archives that need
    to be cleaned. Return 1 if files are found'''

    if os.path.exists('var/cache/apt/archives'):
        skip = ['partial', 'lock']
        count = len([x for x in os.listdir("var/cache/apt/archives/")
                    if x not in skip])

        # The apt cache should be cleaned
        if count > 0:
            print("PROBLEM:  apt-get cache should be cleaned.")
            print("RESOLVE:  sudo apt-get clean")
            returncode = 1

    return returncode


# Debian Cache
if os_base in ["debian", "ubuntu"]:

    if os.path.exists("var/cache"):
        returncode = check_cache(returncode)

    if os.path.exists("var/cache/apt/archives"):
        returncode = check_apt(returncode)

sys.exit(returncode)



'''

test_default_cache.py: Testing default cache function

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import os
import re
import sys
sys.path.append('..')  # noqa

from unittest import TestCase
import shutil
import tempfile

VERSION = sys.version_info[0]

print("*** PYTHON VERSION %s CLIENT TESTING START ***" % VERSION)


class TestDefaultCache(TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        os.environ['SINGULARITY_ROOTFS'] = self.tmpdir
        if "SINGULARITY_CACHE_DIR" in os.environ:
            del os.environ['SINGULARITY_CACHE_DIR']

        print("\n---START----------------------------------------")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

        print("---END------------------------------------------")

    def test_get_cache_default(self):
        '''test_run_command tests sending a command to commandline
        using subprocess
        '''
        print("Testing get_cache...")

        # If there is no cache_base, we should get default
        print("Case 1: No cache base in environment returns default")
        from defaults import SINGULARITY_CACHE
        home = os.environ['HOME']
        self.assertEqual("%s/.singularity" % home, SINGULARITY_CACHE)
        self.assertTrue(os.path.exists(SINGULARITY_CACHE))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

'''

api.py: Singularity Hub helper functions for python

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),
                os.path.pardir)))  # noqa
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) # noqa

from shell import (
    parse_image_uri,
    remove_image_uri,
)

from sutils import (
    add_http,
    clean_up,
    get_image_format,
    read_file,
    run_command
)

from base import ApiConnection

from helpers.json.main import ADD

from defaults import (
    SHUB_API_BASE
)

from message import bot
import json
import re

try:
    from urllib import unquote
except Exception:
    from urllib.parse import unquote


# Shub API Class  ----------------------------------------------

class SingularityApiConnection(ApiConnection):

    def __init__(self, **kwargs):
        self.token = None
        self.api_base = SHUB_API_BASE

        if 'image' in kwargs:
            self.load_image(kwargs['image'])

        if 'token' in kwargs:
            self.token = kwargs['token']
        super(SingularityApiConnection, self).__init__(**kwargs)

    def load_image(self, image):
        self.image = parse_image_uri(image=image,
                                     uri='shub://',
                                     default_registry=SHUB_API_BASE,
                                     quiet=True)
        # parse_image_uri may return an empty namespace cause that's allowed
        # with docker://, but not with shub://
        if len(self.image["namespace"]) == 0:
            bot.error("Namespace cannot be empty for shub:// url!")
            sys.exit(1)

    def get_manifest(self):
        '''get_image will return a json object with image metadata
        based on a unique id.

        Parameters
        ==========
        :param image: the image name, either an id
                      or a repo name, tag, etc.

        Returns
        =======
        manifest: a json manifest from the registry
        '''
        # make sure we have a complete url
        registry = add_http(self.image['registry'])
        base = "%s/api/container/%s/%s:%s" % (registry,
                                              self.image['namespace'],
                                              self.image['repo_name'],
                                              self.image['repo_tag'])
        # ------------------------------------------------------
        # If we need to authenticate, will do it here
        # ------------------------------------------------------

        # If the Hub returns 404, the image name is likely wrong
        response = self.get(base, return_response=True)
        if response.code == 404:
            msg = "Cannot find image."
            msg += " Is your capitalization correct?"
            bot.error(msg)
            sys.exit(1)

        try:
            response = response.read().decode('utf-8')
            response = json.loads(response)

        except Exception:
            print("Error getting image manifest using url %s" % base)
            sys.exit(1)

        return response

    def download_image(self,
                       manifest,
                       download_folder=None,
                       extract=False):

        '''

        download_image will download a singularity image from singularity
        hub to a download_folder, named based on the image version (commit id)

        Parameters
        ==========
        :param manifest: the manifest obtained with get_manifest
        :param download_folder: the folder to download to, if None, will be pwd
        :param extract: if True, will extract image to .img and return that.

        Returns
        =======
        image_file: the full path to the downloaded image

        '''
        from defaults import SHUB_CONTAINERNAME

        # Returns just basename, no extension
        image_name = get_image_name(manifest)

        if not bot.is_quiet():
            print("Found image %s:%s" % (manifest['name'],
                                         manifest['branch']))

            print("Downloading image... %s" % image_name)

        url = manifest['image']

        if url is None:
            bot.error("%s is not ready for download" % image_name)
            bot.error("please try when build completes or specify tag.")
            sys.exit(1)

        if download_folder is not None:
            image_name = "%s/%s" % (download_folder, image_name)

        # Download image file atomically, streaming
        image_file = self.download_atomically(url=url,
                                              file_name=image_name,
                                              show_progress=True)

        # Compressed ext3 images need extraction
        image_type = get_image_format(image_file)
        extension = "simg"

        if image_type == "GZIP" or extract is True:

            extension = "img"
            if not image_file.endswith('.gz'):
                os.rename(image_file, "%s.gz" % image_file)
                image_file = "%s.gz" % image_file

            if not bot.is_quiet():
                print("Decompressing %s" % image_file)

            output = run_command(['gzip', '-d', '-f', image_file])
            image_file = image_file.replace('.gz', '')

            # Any error in extraction (return code not 0) will return None
            if output is None:
                bot.error('Error extracting image, cleaning up.')
                clean_up([image_file, "%s.gz" % image_file])

        # If the user has provided a default name, be true to it
        if SHUB_CONTAINERNAME is not None:
            folder = os.path.dirname(image_file)
            image_name = "%s/%s" % (folder, SHUB_CONTAINERNAME)
            os.rename(image_file, image_name)
            image_file = image_name

        # Otherwise rename to have extension matching image type
        else:
            image_name = "%s.%s" % (image_file, extension)
            os.rename(image_file, image_name)
            image_file = image_name

        return image_file


# Various Helpers -----------------------------------------------
def get_image_name(manifest):
    '''return the image name for a manifest. Does not return extension, and
       does not account for a custom name provided by the user.
       :param manifest: the image manifest with 'image'
                     as key with download link
    '''
    from defaults import (SHUB_CONTAINERNAME,
                          SHUB_NAMEBYCOMMIT,
                          SHUB_NAMEBYHASH)

    if SHUB_CONTAINERNAME is not None:
        return SHUB_CONTAINERNAME

    # Second preference goes to commit
    elif SHUB_NAMEBYCOMMIT is not None:
        if manifest.get("commit") is not None:
            return manifest['commit']
        elif manifest['version'] is not None:
            return manifest['version']

    elif SHUB_NAMEBYHASH is not None:
        image_url = os.path.basename(unquote(manifest['image']))
        image_name = re.findall('([-\w]+\.(?:img|simg|img.gz))', image_url)
        if len(image_name) > 0:
            return image_name[0].split('.')[0]

    return get_default_name(manifest)


def get_default_name(manifest, source="Hub"):
    ''' get manifest name returns the default name that is discovered
        via the image manifest. This is the fallback option in the case that
        the user doesn't ask to name by commit, hash, or custom name
    '''

    # Singularity Hub v2.0
    if "tag" in manifest and "branch" in manifest:
        version = "%s-%s" % (manifest["branch"], manifest["tag"])

    # Singularity Registry
    elif "tag" in manifest:
        version = manifest["tag"]

    # Singularity Hub v1.0
    else:
        version = manifest['branch']

    # Remove slashes
    version = version.replace('/', '-')

    # sregistry images store collection/name separately
    name = manifest['name']
    if 'frozen' in manifest:
        source = "Registry"
        name = '%s-%s' % (manifest['collection'], name)
    image_name = "%s-%s" % (name.replace('/', '-'), version)

    if not bot.is_quiet():
        print("Singularity %s Image: %s" % (source, image_name))
    return image_name


def extract_metadata(manifest, labelfile=None, prefix=None):
    '''extract_metadata will write a file of metadata from shub
    :param manifest: the manifest to use
    '''

    if prefix is None:
        prefix = ""
    prefix = prefix.upper()

    source = 'Hub'
    if 'frozen' in manifest:
        source = 'Registry'

    metadata = manifest.copy()
    remove_fields = ['files', 'spec', 'metrics']
    for remove_field in remove_fields:
        if remove_field in metadata:
            del metadata[remove_field]

    if labelfile is not None:
        for key, value in metadata.items():
            key = "%s%s" % (prefix, key)
            value = ADD(key=key,
                        value=value,
                        jsonfile=labelfile,
                        force=True)

        bot.verbose("Saving Singularity %s metadata to %s" % (source,
                                                              labelfile))
    return metadata



#!/usr/bin/env python
#
# Copyright (c) 2017, SingularityWare, LLC. All rights reserved.
# Copyright (c) 2017, Vanessa Sochat. All rights reserved.
#
# See the COPYRIGHT.md file at the top-level directory of this
# distribution and at https://github.com/singularityware/singularity
#
# This file is part of the Singularity Linux container project.
# It is subject to the license terms in the LICENSE.md file
# found in the top-level directory of this distribution and at
# https://github.com/singularityware/singularity. No part
# of Singularity, including this file, may be copied, modified,
# propagated, or distributed except according to the terms
# contained in the LICENSE.md file.
#
# Yell at the user for bad practices from Docker containers

import sys
import os

base = os.environ["SINGULARITY_ROOTFS"]
os.chdir(base)

returncode = 0

if os.geteuid() != 0:
    print("You must run this test as sudo, skipping")
    sys.exit(returncode)

# Apt-get cache
skip = ['.profile', '.bashrc', '.tcshc',
        '.cshrc', '.bash_history',
        '.bash_profile']

root = [x for x in os.listdir('root') if x not in skip]

# The user should not put content in root!
if len(root) > 0:
    print("PROBLEM:  You should not save content in roots home.")
    print("RESOLVE:  Install to /opt or /usr/local")
    print("\n".join(root))
    returncode = 1

sys.exit(returncode)



'''

templates.py: template data structures for Singularity python
Copyright (c) 2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

from message import bot


def get_template(template_name):
    '''get_template will return a default template for some function in
    Singularity Python. This is to reduce redundancy if data structures
    are used multiple times, etc. If there is no template, None is returned.
    '''
    template_name = template_name.lower()
    templates = dict()

    templates['tarinfo'] = {"gid": 0,
                            "uid": 0,
                            "uname": "root",
                            "gname": "root",
                            "mode": 493}

    if template_name in templates:
        bot.debug("Found template for %s" % (template_name))
        return templates[template_name]
    else:
        bot.warning("Cannot find template %s" % (template_name))
    return None



'''
shell.py: General shell parsing functions for Singularity in Python

Copyright (c) 2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import sys
import os

from message import bot

from defaults import (
    DOCKER_API_BASE,
    NAMESPACE,
    CUSTOM_REGISTRY,
    CUSTOM_NAMESPACE,
    TAG
)

import json
import re
import os


def get_image_uri(image, quiet=False):
    '''get_image_uri will parse a uri sent from Singularity
       to determine if it's  singularity (shub://)
       or docker (docker://)
    :param image: the complete image uri (example: docker://ubuntu:latest
    '''
    image_uri = None
    image = image.replace(' ', '')
    match = re.findall('^[A-Za-z0-9-]+[:]//', image)

    if len(match) == 0:
        if not quiet:
            bot.warning("Could not detect any uri in %s" % image)
    else:
        image_uri = match[0].lower()
        if not quiet:
            bot.debug("Found uri %s" % (image_uri))
    return image_uri


def remove_image_uri(image, image_uri=None, quiet=False):
    '''remove_image_uri will return just the image name.
    ::note this will also remove all spaces from the uri.
    '''
    if image_uri is None:
        image_uri = get_image_uri(image, quiet=quiet)

    image = image.replace(' ', '')

    if image_uri is not None:
        image = image.replace(image_uri, '')
    return image

# Quick python regex syntax reference for referring back to matched groups:
# (?:stuff) matches stuff, just like (stuff) but won't appear in m.groups()
# (?P<name>stuff) matches stuff, and can be retrieved by m.group('name')

# Regular expression used to parse docker:// uris, which have slightly
# different rules than shub://, namely:
# - registry must include :port or a . in the name (e.g. docker.io)
# - namespace is completely optional
_docker_uri_re = re.compile(
    # Optionally match a registry if it contains a '.' or a ':',
    # may contain any character but '/' or '@', and terminated by '/'
    # Note the use of (?:) to make sure the group "registry" does
    # not contain the / at the end. Also, registry includes the port
    "(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
    # Optionally match a namespace, matched as any characters but
    # ':' or '/', followed by '/'. Note the match will include the final /
    "(?P<namespace>(?:[^:@/]+/)+)?"
    # Match a repo name, mandatory. Any character but ':' or '/'
    "(?P<repo>[^:@/]+)"
    # Match :tag (optional)
    "(?::(?P<tag>[^:@]+))?"
    # Match @digest (optional)
    "(?:@(?P<version>.+))?"
    # we need to match the whole string, make sure there's no leftover
    "$"
    )


# Reduced regex, matches registry:port/repo or registry.com/repo
# (registry must include : or .) with optional tag or version.
# Also matches repo only, e.g. reponame[:tag|@version].
# This is tried before _default_uri_re, so that namespace/repo takes
# precedence over registry/repo.
_reduced_uri_no_ns_re = re.compile(
    # match a registry, optional, may include a : or ., but not a @
    "(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
    # Match a repo name, mandatory. Any character but ':' or '/'
    "(?P<repo>[^:@/]+)"
    # Match :tag (optional)
    "(?::(?P<tag>[^:@]+))?"
    # Match @digest (optional)
    "(?:@(?P<version>.+))?"
    # we need to match the whole string, make sure there's no leftover
    "$"
    # dummy group that will never match, but will add a 'namespace' entry
    "(?P<namespace>.)?"
    )

# Regular expression used to parse other images and uris (e.g. shub://)
# This version expects a namespace, and won't match otherwise
# but the registry is optional. In cases like a/b/c, the first component
# is taken to be the registry in this regex.
_default_uri_re = re.compile(
    # must be either registry[:port]/namespace[/more/namespaces]/repo
    # or namespace/repo, at least one namespace is required
    # Match registry, if specified
    "(?:(?P<registry>[^/@]+)/)?"
    # Match a namespace, matched as any characters but
    # ':' or '@', ended by a '/'. Note the match will include the final /
    "(?P<namespace>(?:[^:@/]+/)+)"
    # Match a repo name, mandatory. Any character but ':' or '/'
    "(?P<repo>[^:@/]+)"
    # Match :tag (optional)
    "(?::(?P<tag>[^:@]+))?"
    # Match @digest (optional)
    "(?:@(?P<version>.+))?"
    # we need to match the whole string, make sure there's no leftover
    "$"
    )

def parse_image_uri(image,
                    uri=None,
                    quiet=False,
                    default_registry=None,
                    default_namespace=None,
                    default_tag=None):
    '''parse_image_uri will parse a docker or shub uri and return
    a json structure with a registry, repo name, tag, namespace and version.
    Tag and version are optional, namespace is optional for docker:// uris.
    URIs are of this general form:
        myuri://[registry.com:port/][namespace/nested/]repo[:tag][@version]
    (parts in [] are optional)
    Parsing rules are slightly different if the uri is a docker:// uri:
    - registry must include a :port or a . in its name, else will be parsed
      as a namespace. For non-docker uris, instead, if there are three or
      more parts separated by /, the first one is taken to be the registry
    - namespace can be empty if a registry is specified, else default
      namespace will be used (e.g. docker://registry.com/repo:tag).
      For non-docker uris, namespace cannot be empty and default will be used

    :param image: the string provided on command line for
                  the image name, eg: ubuntu:latest or
                  docker://local.registry/busybox@12345
    :param uri: the uri type (eg, docker://), default autodetects
    ::note uri is maintained as variable so we have some control over allowed
    :param quiet: If True, don't show verbose messages with the parsed values
    :default_registry: Which registry to use if image doesn't contain one.
                       if None, use defaults.REGISTRY
    :default_namespace: Which namespace to use if image doesn't contain one.
                       if None, use defaults.NAMESPACE. Also, check out the
                       note above about docker:// and empty namespaces.
    :default_tag: Which tag to use if image doesn't contain one.
                       if None, use defaults.REPO_TAG
    :returns parsed: a json structure with repo_name, repo_tag, and namespace
    '''

    # Default to most common use case, Docker
    if default_registry is None:
        default_registry = DOCKER_API_BASE

    if default_namespace is None:
        default_namespace = NAMESPACE

    if default_tag is None:
        default_tag = TAG

    # if user gave custom registry / namespace, make them the default
    if CUSTOM_NAMESPACE is not None:
        default_namespace = CUSTOM_NAMESPACE

    if CUSTOM_REGISTRY is not None:
        default_registry = CUSTOM_REGISTRY

    # candidate regex for matching, in order of preference
    uri_regexes = [ _reduced_uri_no_ns_re,
                    _default_uri_re ]

    # Be absolutely sure there are no comments
    image = image.split('#')[0]

    if not uri:
        uri = get_image_uri(image, quiet=True)

    # docker images require slightly different rules
    if uri == "docker://":
        uri_regexes = [ _docker_uri_re ]

    image = remove_image_uri(image, uri)

    for r in uri_regexes:
        match = r.match(image)
        if match:
            break

    if not match:
        bot.error('Could not parse image "%s"! Exiting.' % image)
        sys.exit(1)

    registry = match.group('registry')
    namespace = match.group('namespace')
    repo_name = match.group('repo')
    repo_tag = match.group('tag')
    version = match.group('version')

    if namespace:
        # strip trailing /
        namespace = namespace.rstrip('/')

    # repo_name is required and enforced by the re (in theory)
    # if not provided, re should not match
    assert(repo_name)

    # replace empty fields with defaults
    if not namespace:
        # for docker, if a registry was specified, don't
        # inject a namespace in between
        if registry and uri == "docker://":
            namespace = ""
        else:
            namespace = default_namespace
    if not registry:
        registry = default_registry
    if not repo_tag:
        repo_tag = default_tag
    # version is not mandatory, don't check that


    if not quiet:
        bot.verbose("Registry: %s" % registry)
        bot.verbose("Namespace: %s" % namespace)
        bot.verbose("Repo Name: %s" % repo_name)
        bot.verbose("Repo Tag: %s" % repo_tag)
        bot.verbose("Version: %s" % version)

    parsed = {'registry': registry,
              'namespace': namespace,
              'repo_name': repo_name,
              'repo_tag': repo_tag,
              'version': version }

    return parsed



#!/usr/bin/env python

'''

inspect.py: python helper for Singularity inspect

ENVIRONMENTAL VARIABLES that are used for this executable:

SINGULARITY_MOUNTPOINT
SINGULARITY_INSPECT_LABELS
SINGULARITY_INSPECT_DEFFILE
SINGULARITY_INSPECT_RUNSCRIPT
SINGULARITY_INSPECT_TEST
SINGULARITY_INSPECT_ENVIRONMENT
SINGULARITY_APPNAME

Copyright (c) 2017, Vanessa Sochat. All rights reserved.

'''

import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),
                                os.path.pardir,
                                os.path.pardir)))  # noqa

from helpers.json.main import INSPECT
from message import bot
from defaults import getenv


def main():
    '''this function will run the main inspect function
    '''
    labels = getenv("SINGULARITY_INSPECT_LABELS", None)
    deffile = getenv("SINGULARITY_INSPECT_DEFFILE", None)
    runscript = getenv("SINGULARITY_INSPECT_RUNSCRIPT", None)
    test = getenv("SINGULARITY_INSPECT_TEST", None)
    environment = getenv("SINGULARITY_INSPECT_ENVIRONMENT", None)
    helpfile = getenv("SINGULARITY_INSPECT_HELP", None)
    structured = getenv("SINGULARITY_PRINT_STRUCTURED", None)
    app = getenv("SINGULARITY_APPNAME", None)

    pretty_print = True
    if structured is not None:
        pretty_print = False

    INSPECT(inspect_app=app,
            inspect_labels=labels,
            inspect_def=deffile,
            inspect_runscript=runscript,
            inspect_test=test,
            inspect_help=helpfile,
            inspect_env=environment,
            pretty_print=pretty_print)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

'''

import.py: python helper for Singularity import


ENVIRONMENTAL VARIABLES that are required for this executable:

    SINGULARITY_CONTAINER
    SINGULARITY_CONTENTS


Given that SINGULARITY_ROOTFS is defined, a full import is done that includes
environment, labels, and extraction of layers. If SINGULARITY_ROOTFS is not
defined, then SINGULARITY_CONTENTS must be defined, which returns a list
of layer contents.

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.


'''

import sys
import os

from shell import (
    get_image_uri,
    remove_image_uri
)

from defaults import getenv
from message import bot


def main():
    '''this function will run a docker import, returning a list of layers
    and environmental variables and metadata to the metadata base
    '''

    container = getenv("SINGULARITY_CONTAINER", required=True)
    image_uri = get_image_uri(container, quiet=True)
    container = remove_image_uri(container)

    ########################################################################
    # Docker Image #########################################################
    ########################################################################

    if image_uri == "docker://":

        bot.debug("\n*** STARTING DOCKER IMPORT PYTHON  ****")

        from sutils import basic_auth_header
        from defaults import LAYERFILE

        bot.debug("Docker layers and metadata will be written to: %s"
                  % (LAYERFILE))

        username = getenv("SINGULARITY_DOCKER_USERNAME")
        password = getenv("SINGULARITY_DOCKER_PASSWORD",
                          silent=True)

        auth = None
        if username is not None and password is not None:
            auth = basic_auth_header(username, password)

        from docker.main import IMPORT

        manifest = IMPORT(auth=auth,
                          image=container,
                          layerfile=LAYERFILE)

    ########################################################################
    # Singularity Hub ######################################################
    ########################################################################

    elif image_uri == "shub://":

        bot.debug("\n*** STARTING SINGULARITY HUB IMPORT PYTHON  ****")

        from defaults import LAYERFILE, LABELFILE
        from shub.main import IMPORT
        IMPORT(image=container,
               layerfile=LAYERFILE,
               labelfile=LABELFILE)

    else:
        bot.error("uri %s is not supported for import. Exiting."
                  % (image_uri))
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

'''

pull.py: general "pull" wrapper for Singularity Hub command line tool.
         Currently, only supported endpoint is shub://

ENVIRONMENTAL VARIABLES that are found for this executable:


   SINGULARITY_CONTAINER: container name: shub://vsoch/singularity-images
   SINGULARITY_PULLFOLDER: location to pull image to
   SINGULARITY_METADATA_DIR: if defined, write paths to files here


Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import sys

from shell import (
    get_image_uri,
    remove_image_uri
)

from message import bot
import sys


def main():
    '''main is a wrapper for the client to hand the parser
    to the executable functions
    This makes it possible to set up a parser in test cases
    '''
    bot.debug("\n*** STARTING SINGULARITY PYTHON PULL ****")
    from defaults import LAYERFILE, DISABLE_CACHE, getenv

    # What image is the user asking for?
    container = getenv("SINGULARITY_CONTAINER", required=True)
    pull_folder = getenv("SINGULARITY_PULLFOLDER")

    image_uri = get_image_uri(container)
    container = remove_image_uri(container, quiet=True)

    if image_uri == "shub://":

        from shub.main import PULL
        manifest = PULL(image=container,
                        download_folder=pull_folder,
                        layerfile=LAYERFILE)

    else:
        bot.error("uri %s is not supported for pull. Exiting." % (image_uri))
        sys.exit(1)


if __name__ == '__main__':
    main()



'''

test_json.py: Singularity Hub testing functions for Singularity in Python

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import os
import sys
sys.path.append('..')  # noqa

from unittest import TestCase
import shutil
import tempfile

from subprocess import (
    Popen,
    PIPE,
    STDOUT
)

VERSION = sys.version_info[0]

print("*** PYTHON VERSION %s UTIL HELPERS TESTING START ***" % VERSION)


class TestJson(TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.here = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                 os.path.pardir))
        self.file = "%s/meatballs.json" % self.tmpdir

        print("\n---START----------------------------------------")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        print("---END------------------------------------------")

    def format_keyname(self):
        '''test the function to format the key name
        '''
        from helpers.json.main import format_keyname
        print("Testing formatting of key function.")

        print('Case 1: Testing that key returns all caps')
        key = format_keyname('dry_meatball')
        self.assertEqual(key, 'DRY_MEATBALL')

        print('Case 2: Testing that key replaced invalid characters with _')
        key = format_keyname('big!!meatball)#$FTW')
        self.assertEqual(key, 'DRY_MEATBALL_FTW')
        key = format_keyname('ugly-meatball')
        self.assertEqual(key, 'UGLY_MEATBALL')

    def test_get(self):
        '''test_get will test the get function
        '''
        print('Testing json GET')

        print('Case 1: Get exiting key')
        from sutils import write_json
        write_json({"PASTA": "rigatoni!"}, self.file)
        self.assertTrue(os.path.exists(self.file))

        script_path = "%s/helpers/json/get.py" % self.here
        if VERSION == 2:
            testing_command = ["python2",
                               script_path,
                               '--key',
                               'PASTA',
                               '--file',
                               self.file]
        else:
            testing_command = ["python3",
                               script_path,
                               '--key',
                               'PASTA',
                               '--file',
                               self.file]

        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertEqual(result['return_code'], 0)

        output = result['message']
        if isinstance(output, bytes):
            output = output.decode(encoding='UTF-8')
        self.assertEqual('rigatoni!',
                         output.strip('\n').split('\n')[-1])

        print('Case 2: Get non-existing key exits')
        if VERSION == 2:
            testing_command = ["python2", script_path, '--key',
                               'LASAGNA', '--file', self.file]
        else:
            testing_command = ["python3", script_path, '--key',
                               'LASAGNA', '--file', self.file]

        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertEqual(result['return_code'], 1)

    def test_add_delete(self):
        '''test_add_delete will test the add and delete functions
        '''
        print('Testing json ADD')
        from sutils import write_json, read_json

        print('Case 1: Adding to new file, force not needed')
        self.assertTrue(os.path.exists(self.file) is False)

        script_path = "%s/helpers/json/add.py" % self.here
        if VERSION == 2:
            testing_command = ["python2", script_path, '--key', 'LEGO',
                               '--value', 'RED', '--file', self.file]
        else:
            testing_command = ["python3", script_path, '--key', 'LEGO',
                               '--value', 'RED', '--file', self.file]

        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)

        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertEqual(result['return_code'], 0)
        self.assertTrue(os.path.exists(self.file))

        # Check the contents
        contents = read_json(self.file)
        self.assertTrue('LEGO' in contents)
        self.assertTrue(contents['LEGO'] == 'RED')

        print('Case 2: Adding to existing key without force should error.')
        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertEqual(result['return_code'], 1)

        print('Case 3: Adding to existing key with force should work.')
        if VERSION == 2:
            testing_command = ["python2", script_path, '--key', 'LEGO',
                               '--value', 'BLUE', '--file', self.file, '-f']
        else:
            testing_command = ["python3", script_path, '--key', 'LEGO',
                               '--value', 'BLUE', '--file', self.file, '-f']

        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertEqual(result['return_code'], 0)

        # Check the updated contents
        contents = read_json(self.file)
        self.assertTrue('LEGO' in contents)
        self.assertTrue(contents['LEGO'] == 'BLUE')

        if VERSION == 2:
            testing_command = ["python2", script_path, '--key', 'PASTA',
                               '--value', 'rigatoni!', '--file', self.file]
        else:
            testing_command = ["python3", script_path, '--key', 'PASTA',
                               '--value', 'rigatoni!', '--file', self.file]

        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}

        print('Case 4: Deleting key from file')
        script_path = "%s/helpers/json/delete.py" % self.here
        if VERSION == 2:
            testing_command = ["python2", script_path, '--key',
                               'LEGO', '--file', self.file]
        else:
            testing_command = ["python3", script_path,
                               '--key', 'LEGO', '--file', self.file]

        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertEqual(result['return_code'], 0)

        # Check the key was deleted contents
        contents = read_json(self.file)
        self.assertTrue('LEGO' not in contents)

        print('Case 5: Checking that empty file is removed.')
        if VERSION == 2:
            testing_command = ["python2", script_path, '--key',
                               'PASTA', '--file', self.file]
        else:
            testing_command = ["python3", script_path, '--key',
                               'PASTA', '--file', self.file]

        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertTrue(os.path.exists(self.file) is False)

    def test_dump(self):
        '''test_add_delete will test the add and delete functions
        '''
        print('Testing json DUMP')
        from sutils import write_json, read_json

        print('Case 1: Dumping file.')

        jsondump = {'HELLO': 'KITTY',
                    'BATZ': 'MARU',
                    'MY': 'MELODY'}
        write_json(jsondump, self.file)
        self.assertTrue(os.path.exists(self.file))

        script_path = "%s/helpers/json/dump.py" % self.here
        if VERSION == 2:
            testing_command = ["python2", script_path,
                               '--file', self.file]
        else:
            testing_command = ["python3", script_path,
                               '--file', self.file]

        output = Popen(testing_command,
                       stderr=PIPE,
                       stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertEqual(result['return_code'], 0)

        output = result['message']
        if isinstance(output, bytes):
            output = output.decode(encoding='UTF-8')

        dump = ['HELLO:"KITTY"', 'BATZ:"MARU"', 'MY:"MELODY"']
        result = output.strip('\n').split('\n')[-3:]
        for res in result:
            self.assertTrue(res in dump)


if __name__ == '__main__':
    unittest.main()



'''

api.py: Docker helper functions for Singularity in Python

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import sys
import math
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),
                                os.path.pardir)))  # noqa
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) # noqa

from base import ApiConnection

from sutils import (
    add_http,
    get_cache,
    change_tar_permissions,
    create_tar,
    write_singularity_infos
)

from defaults import (
    DOCKER_API_BASE,
    DOCKER_API_VERSION,
    DOCKER_ARCHITECTURE,
    DOCKER_OS,
    DOCKER_NUMBER,
    DOCKER_PREFIX,
    ENV_BASE,
    LABELFILE,
    METADATA_FOLDER_NAME,
    RUNSCRIPT_COMMAND_ASIS
)

from helpers.json.main import ADD
from message import bot
from shell import parse_image_uri
from templates import get_template

import json
import re
import tempfile
try:
    from urllib.error import HTTPError
except ImportError:
    from urllib2 import HTTPError


# Docker API Class  ---------------------------------------

class DockerApiConnection(ApiConnection):

    def __init__(self, **kwargs):
        self.auth = None
        self.token = None
        self.token_url = None
        self.schemaVersion = None
        self.reverseLayers = False
        self.api_base = DOCKER_API_BASE
        self.api_version = DOCKER_API_VERSION
        self.manifest = None
        self.manifestv1 = None

        if 'auth' in kwargs:
            self.auth = kwargs['auth']
        if 'token' in kwargs:
            self.token = kwargs['token']
        super(DockerApiConnection, self).__init__(**kwargs)
        if 'image' in kwargs:
            self.load_image(kwargs['image'])

    def assemble_uri(self, sep=None):
        '''re-assemble the image uri, for components defined.
        '''
        if sep is None:
            sep = "-"

        image_uri = "%s%s%s" % (self.registry, sep, self.repo_name)

        if self.version is not None:
            image_uri = "%s@%s" % (image_uri, self.version)
        else:
            image_uri = "%s:%s" % (image_uri, self.repo_tag)

        return image_uri

    def _init_headers(self):
        # specify wanting version 2 schema
        # meaning the correct order of digests
        # returned (base to child)
        return {"Accept": 'application/vnd.docker.distribution.manifest.v2+json,application/vnd.docker.distribution.manifest.list.v2+json,application/json',  # noqa
               'Content-Type': 'application/json; charset=utf-8'}  # noqa

    def check_errors(self, response, exit=True):
        '''take a response with errors key,
        iterate through errors in expected format and exit upon completion
        '''
        if "errors" in response:
            for error in response['errors']:
                bot.error("%s: %s" % (error['code'],
                                      error['message']))
                if error['code'] == "UNAUTHORIZED":
                    msg = "Check existence, naming, and permissions"
                    bot.error(msg)
            if exit:
                sys.exit(1)
        return response

    def load_image(self, image):
        '''load_image parses the image uri, and loads the
        different image parameters into the client.
        The image should be a docker uri (eg docker://)
        or name of docker image'''

        image = parse_image_uri(image=image, uri="docker://")
        if len(image['namespace']) > 0:
            self.repo_name = "%s/%s" % (image['namespace'], image['repo_name'])
        else:
            self.repo_name = image['repo_name']

        self.repo_tag = image['repo_tag']
        self.version = image['version']
        self.registry = image['registry']
        self.update_token()

    def update_token(self, response=None, auth=None):
        '''update_token uses HTTP basic authentication to get a token for
        Docker registry API V2 operations. We get here if a 401 is
        returned for a request.
        https://docs.docker.com/registry/spec/auth/token/
        '''
        if self.token_url is None:

            if response is None:
                response = self.get_tags(return_response=True)

            if not isinstance(response, HTTPError):
                bot.verbose3('Response on obtaining token is None.')
                return None

            not_asking_auth = "Www-Authenticate" not in response.headers
            if response.code != 401 or not_asking_auth:
                bot.error("Authentication error, exiting.")
                sys.exit(1)

            challenge = response.headers["Www-Authenticate"]
            regexp = '^Bearer\s+realm="(.+)",service="(.+)",scope="(.+)",?'
            match = re.match(regexp, challenge)

            if not match:
                bot.error("Unrecognized authentication challenge, exiting.")
                sys.exit(1)

            realm = match.group(1)
            service = match.group(2)
            scope = match.group(3).split(',')[0]

            self.token_url = ("%s?service=%s&expires_in=9000&scope=%s"
                              % (realm, service, scope))

        headers = dict()

        # First priority comes to auth supplied directly to function
        if auth is not None:
            headers.update(auth)

        # Second priority is default if supplied at init
        elif self.auth is not None:
            headers.update(self.auth)

        response = self.get(self.token_url,
                            default_headers=False,
                            headers=headers)

        try:
            token = json.loads(response)["token"]
            token = {"Authorization": "Bearer %s" % token}
            self.token = token
            self.update_headers(token)

        except Exception:
            bot.error("Error getting token for repository %s, exiting."
                      % self.repo_name)
            sys.exit(1)

    def get_images(self, manifest=None):
        '''get_images will return a list of layers from a manifest.
        The function is intended to work with both version
        1 and 2 of the schema. All layers (including redundant)
        are returned.

        For version 1 manifests: extraction is reversed

        :param manifest: the manifest to read_layers from
        '''
        if manifest is None:
            self.update_manifests()
            manifest = self.manifest

        digests = []
        layer_key = 'layers'
        digest_key = 'digest'

        # Docker manifest-v2-2.md#image-manifest
        if 'layers' in manifest:
            bot.debug('Image manifest version 2.2 found.')

        # Docker manifest-v2-1.md#example-manifest  # noqa
        elif 'fsLayers' in manifest:
            layer_key = 'fsLayers'
            digest_key = 'blobSum'
            bot.debug('Image manifest version 2.1 found.')

        else:
            msg = "Improperly formed manifest, "
            msg += "layers, manifests, or fsLayers must be present"
            bot.error(msg)
            sys.exit(1)

        for layer in manifest[layer_key]:
            if digest_key in layer:
                bot.debug("Adding digest %s" % layer[digest_key])
                digests.append(layer[digest_key])

        # Reverse layer order for manifest version 1.0
        if self.reverseLayers is True:
            message = 'v%s manifest, reversing layers' % self.schemaVersion
            bot.debug(message)
            digests.reverse()

        return digests

    def get_tags(self, return_response=False):
        '''get_tags will return the tags for a repo using the
        Docker Version 2.0 Registry API
        '''
        registry = self.registry
        if registry is None:
            registry = self.api_base

        registry = add_http(registry)  # make sure we have a complete url

        base = "%s/%s/%s/tags/list" % (registry,
                                       self.api_version,
                                       self.repo_name)

        bot.verbose("Obtaining tags: %s" % base)

        # We use get_tags for a testing endpoint in update_token
        response = self.get(base,
                            return_response=return_response)

        if return_response:
            return response

        try:
            response = json.loads(response)
            return response['tags']
        except Exception:
            bot.error("Error obtaining tags: %s" % base)
            sys.exit(1)

    def get_manifest(self, old_version=False, version=None):
        '''get_manifest should return an image manifest
        for a particular repo and tag.  The image details
        are extracted when the client is generated.
        :param old_version: return version 1
                            (for cmd/entrypoint), default False
        '''
        registry = self.registry
        if registry is None:
            registry = self.api_base

        # make sure we have a complete url
        registry = add_http(registry)

        base = "%s/%s/%s/manifests" % (registry,
                                       self.api_version,
                                       self.repo_name)

        # First priority given to calling function
        if version is not None:
            base = "%s/%s" % (base, version)

        elif self.version is not None:
            base = "%s/%s" % (base, self.version)

        else:
            base = "%s/%s" % (base, self.repo_tag)
        bot.verbose("Obtaining manifest: %s" % base)

        headers = self.headers.copy()
        if old_version is True:
            headers['Accept'] = 'application/json'

        response = self.get(base, headers=headers)

        try:
            response = json.loads(response)

        except Exception:

            # If the call fails, give the user a list of acceptable tags
            tags = self.get_tags()
            print("\n".join(tags))
            repo_uri = "%s:%s" % (self.repo_name,
                                  self.repo_tag)

            bot.error("Error getting manifest for %s, exiting." % repo_uri)
            sys.exit(1)

        # If we have errors, don't continue
        return self.check_errors(response)

    def update_manifests(self):
        '''update manifests ensures that each of a version1 and version2
        manifest are present
        '''
        bot.debug('Updating manifests.')

        if self.repo_name is None:
            bot.error("Insufficient metadata to get manifest.")
            sys.exit(1)

        # Get full image manifest, using version 2.0 of Docker Registry API
        if self.manifest is None:
            bot.debug('MANIFEST (Primary): not found, making initial call.')
            self.manifest = self.get_manifest()
            # This is the primary manifest schema version, determines if we
            # need to reverse layers
            self.schemaVersion = self.manifest['schemaVersion']
            if self.schemaVersion == 1:
                self.reverseLayers = True

        if self.manifestv1 is None:
            bot.debug('MANIFEST (Metadata): not found, making initial call.')
            self.manifestv1 = self.get_manifest(old_version=True)

        # https://docs.docker.com/registry/spec/manifest-v2-2/#manifest-list
        if "manifests" in self.manifest:
            for entry in self.manifest['manifests']:
                if entry['platform']['architecture'] == DOCKER_ARCHITECTURE:
                    if entry['platform']['os'] == DOCKER_OS:
                        digest = entry['digest']
                        bot.debug('Image manifest version 2.2 list found.')
                        bot.debug('Obtaining architecture: %s, OS: %s'
                                  % (DOCKER_ARCHITECTURE, DOCKER_OS))

                        # Obtain specific os, architecture
                        self.manifest = self.get_manifest(version=digest)
                        break

        # If we didn't get a new manifest, fall back to version 1
        if "manifests" in self.manifest:
            self.manifest = self.manifestv1

    def get_layer(self,
                  image_id,
                  download_folder=None,
                  change_perms=False,
                  return_tmp=False):

        '''get_layer will download an image layer (.tar.gz)
        to a specified download folder.
        :param download_folder: if specified, download to folder.
                                Otherwise return response with raw data
        :param change_perms: change permissions additionally
                             (default False to support multiprocessing)
        :param return_tmp: If true, return the temporary file name (and
                             don't rename to the file's final name). Default
                             is False, should be True for multiprocessing
                             that requires extra permission changes
        '''
        registry = self.registry
        if registry is None:
            registry = self.api_base

        # make sure we have a complete url
        registry = add_http(registry)

        # The <name> variable is repo_name
        base = "%s/%s/%s/blobs/%s" % (registry,
                                      self.api_version,
                                      self.repo_name,
                                      image_id)
        bot.verbose("Downloading layers from %s" % base)

        if download_folder is None:
            download_folder = tempfile.mkdtemp()

        download_folder = "%s/%s.tar.gz" % (download_folder, image_id)

        # Update user what we are doing
        bot.debug("Downloading layer %s" % image_id)

        # Step 1: Download the layer atomically
        file_name = "%s.%s" % (download_folder,
                               next(tempfile._get_candidate_names()))
        tar_download = self.download_atomically(url=base,
                                                file_name=file_name)
        bot.debug('Download of raw file (pre permissions fix) is %s'
                  % tar_download)

        # Step 2: Fix Permissions?
        if change_perms:
            tar_download = change_tar_permissions(tar_download)

        if return_tmp is True:
            return tar_download

        try:
            os.rename(tar_download, download_folder)
        except Exception:
            msg = "Cannot untar layer %s," % tar_download
            msg += " was there a problem with download?"
            bot.error(msg)
            sys.exit(1)
        return download_folder

    def get_size(self, add_padding=True, round_up=True, return_mb=True):
        '''get_size will return the image size (must use v.2.0 manifest)
        :add_padding: if true, return reported size * 5
        :round_up: if true, round up to nearest integer
        :return_mb: if true, defaults bytes are converted to MB
        '''
        self.update_manifests()
        size = None
        if "layers" in self.manifest:
            size = 0
            for layer in self.manifest["layers"]:
                if "size" in layer:
                    size += layer['size']

            if add_padding is True:
                size = size * 5

            if return_mb is True:
                size = size / (1024 * 1024)  # 1MB = 1024*1024 bytes

            if round_up is True:
                size = math.ceil(size)
            size = int(size)

        return size

    def get_config(self, spec="Entrypoint", delim=None, old_version=False):
        '''get_config returns a particular spec (default is Entrypoint)
        from a VERSION 1 manifest obtained with get_manifest.
        :param manifest: the manifest obtained from get_manifest
        :param spec: the key of the spec to return, default is "Entrypoint"
        :param delim: Given a list, the delim to use to join the entries.
                      Default is newline
        '''
        manifest = self.get_manifest(old_version=old_version)

        cmd = None

        # Version 1 of the manifest has more detailed metadata
        if old_version:
            if "history" in manifest:
                for entry in manifest['history']:
                    if 'v1Compatibility' in entry:
                        entry = json.loads(entry['v1Compatibility'])
                        if "config" in entry:
                            if spec in entry["config"]:
                                cmd = entry["config"][spec]

            # Standard is to include commands like ['/bin/sh']
            if isinstance(cmd, list):
                if delim is None:
                    delim = "\n"
                cmd = delim.join(cmd)
            bot.verbose("Found Docker config (%s) %s" % (spec, cmd))

        else:
            if "config" in manifest:
                if spec in manifest['config']:
                    cmd = manifest['config'][spec]
        return cmd



'''

test_shub.py: Singularity Hub testing functions for Singularity in Python

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import os
import sys
sys.path.append('..')  # noqa

from unittest import TestCase
from sutils import read_file
from glob import glob
import shutil
import tempfile

VERSION = sys.version_info[0]

print("*** PYTHON VERSION %s SINGULARITY HUB API TESTING START ***" % VERSION)


class TestApi(TestCase):

    def setUp(self):
        from shub.api import SingularityApiConnection
        self.connect = SingularityApiConnection
        self.image = "shub://vsoch/singularity-images"
        self.client = SingularityApiConnection(image=self.image)

        self.user_name = "vsoch"
        self.repo_name = "singularity-images"
        self.tmpdir = tempfile.mkdtemp()
        os.environ['SINGULARITY_ROOTFS'] = self.tmpdir
        os.mkdir('%s/.singularity.d' % self.tmpdir)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_get_manifest(self):
        '''test_get_manifest should return the shub manifest
        '''
        print("Testing singularity hub manifest retrieval")
        print("...case 1: Testing retrieval of singularity-hub manifest")
        manifest = self.client.get_manifest()
        keys = ['version', 'tag', 'branch',
                'name', 'id', 'commit', 'image']
        [self.assertTrue(x in manifest) for x in keys]

    def test_download_image(self):
        '''test_download_image will ensure that
        an image is downloaded to an
        appropriate location (tmpdir) or cache
        '''
        print("Testing singularity hub pull")
        print("...case 1: Specifying a directory downloads to it")
        manifest = self.client.get_manifest()
        image = self.client.download_image(manifest=manifest,
                                           download_folder=self.tmpdir)
        self.assertEqual(os.path.dirname(image), self.tmpdir)

        print("...case 2: Not specifying a directory downloads to PWD")
        os.chdir(self.tmpdir)
        image = self.client.download_image(manifest)
        self.assertEqual(os.getcwd(), self.tmpdir)
        print(image)
        self.assertTrue(image in glob("*"))
        os.remove(image)

    def test_uri(self):
        '''test_uri will make sure that the endpoint returns the equivalent
        image for all different uri options
        '''
        print("Testing singularity hub uris")
        from shub.api import get_image_name
        manifest = self.client.get_manifest()
        image_name = get_image_name(manifest)

        fullname = "%s/%s" % (self.user_name, self.repo_name)
        print("...case 1: ask for image by tag.")
        client = self.connect(image="%s:latest" % fullname)
        manifest = client.get_manifest()
        image_name = get_image_name(manifest)
        self.assertEqual(image_name, get_image_name(manifest))

    def test_get_image_name(self):
        '''test_get_image_name will return the image name from the manifest
        '''
        print("Testing singularity hub image naming")
        from shub.api import get_image_name
        manifest = self.client.get_manifest()
        print("...case 1: return an image name corresponding to repo")
        image_name = get_image_name(manifest)
        self.assertEqual('vsoch-singularity-images-master-latest', image_name)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

'''
utils.py: python helper for singularity command line tool

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

from defaults import (
    SINGULARITY_CACHE,
    DISABLE_HTTPS
)
from message import bot
import datetime
import errno
import hashlib
import json
import os
import subprocess
import stat
import sys
import tempfile
import tarfile
import base64
import re

from io import (
    BytesIO,
    StringIO
)

# Python less than version 3 must import OSError
if sys.version_info[0] < 3:
    from exceptions import OSError


####################################################################
# HTTP OPERATIONS ##################################################
####################################################################

def add_http(url, use_https=True):
    '''add_http will add a http / https prefix to a url,
    in case the user didn't specify
    :param url: the url to add the prefix to
    :param use_https: should we default to https? default is True
    '''
    scheme = "https://"
    if use_https is False or DISABLE_HTTPS is True:
        scheme = "http://"

    # remove scheme from url
    # urlparse is buggy in Python 2.6
    # https://bugs.python.org/issue754016, use regex instead
    parsed = re.sub('.*//', '', url)

    return "%s%s" % (scheme, parsed.rstrip('/'))


def basic_auth_header(username, password):
    '''basic_auth_header will return a base64 encoded header object to
    generate a token
    :param username: the username
    :param password: the password
    '''
    s = "%s:%s" % (username, password)
    if sys.version_info[0] >= 3:
        s = bytes(s, 'utf-8')
        credentials = base64.b64encode(s).decode('utf-8')
    else:
        credentials = base64.b64encode(s)
    auth = {"Authorization": "Basic %s" % credentials}
    return auth


####################################################################
# COMMAND LINE OPERATIONS ##########################################
####################################################################


def run_command(cmd, env=None, quiet=False):
    '''run_command uses subprocess to send a command to the terminal.
    :param cmd: the command to send, should be a list for subprocess
    :param env: an optional environment to include, a dictionary of key/values
    '''
    try:
        if quiet is False:
            bot.verbose2("Running command %s with subprocess" % " ".join(cmd))
        if env is None:
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        else:
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env)
    except OSError as error:
        bot.error("Error with subprocess: %s, returning None" % error)
        return None

    output = process.communicate()[0]
    if process.returncode != 0:
        return None

    return output


def clean_up(files):
    '''clean up will delete a list of files, only if they exist
    '''
    if not isinstance(files, list):
        files = [files]

    for f in files:
        if os.path.exists(f):
            bot.verbose3("Cleaning up %s" % f)
            os.remove(f)

############################################################################
# TAR/COMPRESSION ##########################################################
############################################################################


def extract_tar(archive, output_folder):
    '''extract a tar archive to a specified output folder
    :param archive: the archive file to extract
    :param output_folder: the output folder to extract to
    '''
    # If extension is .tar.gz, use -xzf
    args = '-xf'
    if archive.endswith(".tar.gz"):
        args = '-xzf'

    # Just use command line, more succinct.
    command = ["tar", args, archive, "-C", output_folder, "--exclude=dev/*"]
    if not bot.is_quiet():
        print("Extracting %s" % archive)

    return run_command(command)


def create_tar(files, output_folder=None):
    '''create_memory_tar will take a list of files (each a dictionary
    with name, permission, and content) and write the tarfile
    (a sha256 sum name is used) to the output_folder.
    If there is no output folde specified, the
    tar is written to a temporary folder.
    '''
    if output_folder is None:
        output_folder = tempfile.mkdtemp()

    finished_tar = None
    additions = []
    contents = []

    for entity in files:
        info = tarfile.TarInfo(name=entity['name'])
        info.mode = entity['mode']
        info.mtime = int(datetime.datetime.now().strftime('%s'))
        info.uid = entity["uid"]
        info.gid = entity["gid"]
        info.uname = entity["uname"]
        info.gname = entity["gname"]

        # Get size from stringIO write
        filey = StringIO()
        content = None
        try:  # python3
            info.size = filey.write(entity['content'])
            content = BytesIO(entity['content'].encode('utf8'))
        except Exception:  # python2
            info.size = int(filey.write(entity['content'].decode('utf-8')))
            content = BytesIO(entity['content'].encode('utf8'))
        pass

        if content is not None:
            addition = {'content': content,
                        'info': info}
            additions.append(addition)
            contents.append(content)

    # Now generate the sha256 name based on content
    if len(additions) > 0:
        hashy = get_content_hash(contents)
        finished_tar = "%s/sha256:%s.tar.gz" % (output_folder, hashy)

        # Warn the user if it already exists
        if os.path.exists(finished_tar):
            msg = "metadata file %s already exists " % finished_tar
            msg += "will over-write."
            bot.debug(msg)

        # Add all content objects to file
        tar = tarfile.open(finished_tar, "w:gz")
        for a in additions:
            tar.addfile(a["info"], a["content"])
        tar.close()

    else:
        msg = "No contents, environment or labels"
        msg += " for tarfile, will not generate."
        bot.debug(msg)

    return finished_tar


####################################################################
# HASHES/FORMAT ####################################################
####################################################################

def get_content_hash(contents):
    '''get_content_hash will return a hash for a list of content (bytes/other)
    '''
    hasher = hashlib.sha256()
    for content in contents:
        if isinstance(content, BytesIO):
            content = content.getvalue()
        if not isinstance(content, bytes):
            content = bytes(content)
        hasher.update(content)
    return hasher.hexdigest()


def get_image_format(image_file):
    '''
       get image format will use the image-format executable to return the kind
       of file type for the image
       Parameters
       ==========
       image_file: full path to the image file to inspect
       Returns
       =======
       GZIP, DIRECTORY, SQUASHFS, EXT3
    '''
    if image_file.endswith('gz'):
        bot.debug('Found compressed image')
        return "GZIP"

    here = os.path.abspath(os.path.dirname(__file__))
    sbin = here.replace("python", "bin/image-type")
    custom_env = os.environ.copy()
    custom_env["SINGULARITY_MESSAGELEVEL"] = "1"
    image_format = run_command([sbin, image_file], env=custom_env, quiet=True)
    if image_format is not None:
        if isinstance(image_format, bytes):
            image_format = image_format.decode('utf-8')
        image_format = str(image_format).strip('\n')
    bot.debug('Found %s image' % image_format)
    return image_format


####################################################################
# FOLDERS ##########################################################
####################################################################


def get_cache(subfolder=None, quiet=False):
    '''get_cache will return the user's cache for singularity.
    The path returned is generated at the start of the run,
    and returned optionally with a subfolder
    :param subfolder: a subfolder in the cache base to retrieve
    '''

    # Clean up the path and create
    cache_base = clean_path(SINGULARITY_CACHE)

    # Does the user want to get a subfolder in cache base?
    if subfolder is not None:
        cache_base = "%s/%s" % (cache_base, subfolder)

    # Create the cache folder(s), if don't exist
    create_folders(cache_base)

    if not quiet:
        bot.info("Cache folder set to %s" % cache_base)
    return cache_base


def create_folders(path):
    '''create_folders attempts to get the same functionality as mkdir -p
    :param path: the path to create.
    '''
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            bot.error("Error creating path %s, exiting." % path)
            sys.exit(1)


############################################################################
# PERMISSIONS ##############################################################
############################################################################


def has_permission(file_path, permission=None):
    '''has_writability will check if a file has writability using
    bitwise operations. file_path can be a tar member
    :param file_path: the path to the file, or tar member
    :param permission: the stat permission to check for
    is False)
    '''
    if permission is None:
        permission = stat.S_IWUSR
    if isinstance(file_path, tarfile.TarInfo):
        has_permission = file_path.mode & permission
    else:
        st = os.stat(file_path)
        has_permission = st.st_mode & permission
    if has_permission > 0:
        return True
    return False


def change_tar_permissions(tar_file,
                           file_permission=None,
                           folder_permission=None):

    '''change_tar_permissions changes a permission if
    any member in a tarfile file does not have it
    :param file_path the path to the file
    :param file_permission: stat permission to use for files
    :param folder_permission: stat permission to use for folders
    '''
    tar = tarfile.open(tar_file, "r:gz")

    # Owner read, write (o+rw)
    if file_permission is None:
        file_permission = stat.S_IRUSR | stat.S_IWUSR

    # Owner read, write execute (o+rwx)
    if folder_permission is None:
        folder_permission = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR

    # Add owner write permission to all, not symlinks
    members = tar.getmembers()

    if len(members) > 0:

        bot.verbose("Fixing permission for %s" % tar_file)

        # Add all content objects to file
        fd, tmp_tar = tempfile.mkstemp(prefix=("%s.fixperm." % tar_file))
        os.close(fd)
        fixed_tar = tarfile.open(tmp_tar, "w:gz")

        for member in members:

            # add o+rwx for directories
            if member.isdir() and not member.issym():
                member.mode = folder_permission | member.mode
                extracted = tar.extractfile(member)
                fixed_tar.addfile(member, extracted)

            # add o+rw for plain files
            elif member.isfile() and not member.issym():
                member.mode = file_permission | member.mode
                extracted = tar.extractfile(member)
                fixed_tar.addfile(member, extracted)
            else:
                fixed_tar.addfile(member)

        fixed_tar.close()
        tar.close()

        # Rename the fixed tar to be the old name
        os.rename(tmp_tar, tar_file)
    else:
        tar.close()
        bot.warning("Tar file %s is empty, skipping." % tar_file)

    return tar_file

############################################################################
# FILES ####################################################################
############################################################################


def write_file(filename, content, mode="w"):
    '''write_file will open a file, "filename"
    and write content, "content" and properly close the file
    '''
    bot.verbose2("Writing file %s with mode %s." % (filename, mode))
    with open(filename, mode) as filey:
        filey.writelines(content)
    return filename


def write_json(json_obj, filename, mode="w", print_pretty=True):
    '''write_json will (optionally,pretty print) a json object to file
    :param json_obj: the dict to print to json
    :param filename: the output file to write to
    :param pretty_print: if True, will use nicer formatting
    '''
    bot.verbose2("Writing json file %s with mode %s." % (filename, mode))
    with open(filename, mode) as filey:
        if print_pretty is True:
            filey.writelines(print_json(json_obj))
        else:
            filey.writelines(json.dumps(json_obj))
    return filename


def read_json(filename, mode='r'):
    '''read_json reads in a json file and returns
    the data structure as dict.
    '''
    with open(filename, mode) as filey:
        data = json.load(filey)
    return data


def read_file(filename, mode="r", readlines=True):
    '''read_file will open a file, "filename" and
    read content, "content" and properly close the file
    '''
    bot.verbose3("Reading file %s with mode %s." % (filename, mode))
    with open(filename, mode) as filey:
        if readlines:
            content = filey.readlines()
        else:
            content = filey.read()
    return content


def print_json(content, print_console=False):
    '''print_json is intended to pretty print a json
    :param content: the dictionary to print
    :param print_console: if False, return the dump as string (default)
    '''
    if print_console:
        print(json.dumps(content, indent=4, separators=(',', ': ')))
    else:
        return json.dumps(content, indent=4, separators=(',', ': '))


def clean_path(path):
    '''clean_path will canonicalize the path
    :param path: the path to clean
    '''
    return os.path.realpath(path.strip(" "))


def get_fullpath(file_path, required=True):
    '''get_fullpath checks if a file exists, and returns the
    full path to it if it does. If required is true, an error is triggered.
    :param file_path: the path to check
    :param required: is the file required? If True, will exit with error
    '''
    file_path = os.path.abspath(file_path)
    if os.path.exists(file_path):
        return file_path

    # If file is required, we exit
    if required is True:
        bot.error("Cannot find file %s, exiting." % file_path)
        sys.exit(1)

    # If file isn't required and doesn't exist, return None
    bot.warning("Cannot find file %s" % file_path)
    return None


def get_next_infos(base_dir, prefix, start_number, extension):
    '''get_next infos will browse some directory and return
    the next available file
    '''
    output_file = None
    counter = start_number
    found = False

    while not found:
        output_file = "%s/%s-%s%s" % (base_dir,
                                      counter,
                                      prefix,
                                      extension)
        if not os.path.exists(output_file):
            found = True
        counter += 1
    return output_file


def write_singularity_infos(base_dir,
                            prefix,
                            start_number,
                            content,
                            extension=None):

    '''write_singularity_infos will write some metadata object
    to a file in some base, starting at some default number. For example,
    we would want to write dockerN files with docker environment exports to
    some directory ENV_DIR and increase N until we find an available path
    :param base_dir: the directory base to write the file to
    :param prefix: the name of the file prefix (eg, docker)
    :param start_number: the number to start looking for available file at
    :param content: the content to write
    :param extension: the extension to use. If not defined, uses .sh
    '''
    if extension is None:
        extension = ""
    else:
        extension = ".%s" % extension

    # if the base directory doesn't exist, exit with error.
    if not os.path.exists(base_dir):
        msg = "Cannot find required metadata directory"
        msg = "%s %s. Exiting!" % (msg, base_dir)
        bot.warning(msg)
        sys.exit(1)

    # Get the next available number
    output_file = get_next_infos(base_dir,
                                 prefix,
                                 start_number,
                                 extension)
    write_file(output_file, content)
    return output_file



'''

main.py: Docker helper functions for Singularity in Python

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import os
from defaults import (
    INCLUDE_CMD,
    PLUGIN_FIXPERMS
)

from base import MultiProcess

from sutils import (
    get_cache,
    write_file
)

from .api import (
    DockerApiConnection,
)

from .tasks import (
    download_layer,
    change_permissions,
    extract_runscript,
    extract_metadata_tar
)

from message import bot


def SIZE(image, auth=None, contentfile=None):
    '''size is intended to be run before an import, to return to the
    contentfile a list of sizes (one per layer) corresponding
    with the layers that will be downloaded for image
    '''
    bot.debug("Starting Docker SIZE, will get size from manifest")
    bot.verbose("Docker image: %s" % image)
    client = DockerApiConnection(image=image, auth=auth)
    size = client.get_size()
    if contentfile is not None:
        write_file(contentfile, str(size), mode="w")
    return size


def IMPORT(image, auth=None, layerfile=None):
    '''IMPORT is the main script that will obtain docker layers,
    runscript information (either entrypoint or cmd), and environment
    and return a list of tarballs to extract into the image
    :param auth: if needed, an authentication header (default None)
    :param layerfile: The file to write layers to extract into
    '''
    bot.debug("Starting Docker IMPORT, includes env, runscript, and metadata.")
    bot.verbose("Docker image: %s" % image)

    # Does the user want to override default of using ENTRYPOINT?
    if INCLUDE_CMD:
        bot.verbose2("Specified Docker CMD as %runscript.")
    else:
        bot.verbose2("Specified Docker ENTRYPOINT as %runscript.")

    # Input Parsing ----------------------------
    # Parse image name, repo name, and namespace
    client = DockerApiConnection(image=image, auth=auth)

    docker_image_uri = "Docker image path: %s" % client.assemble_uri("/")
    bot.info(docker_image_uri)

    # IMAGE METADATA -------------------------------------------
    # Use Docker Registry API (version 2.0) to get images ids, manifest

    images = client.get_images()

    #  DOWNLOAD LAYERS -------------------------------------------
    # Each is a .tar.gz file, obtained from registry with curl

    # Get the cache (or temporary one) for docker
    cache_base = get_cache(subfolder="docker")
    download_client = MultiProcess()

    # Generate a queue of tasks to run with MultiProcess
    layers = []
    tasks = []
    for ii in range(len(images)):
        image_id = images[ii]
        targz = "%s/%s.tar.gz" % (cache_base, image_id)
        if not os.path.exists(targz):
            tasks.append((client, image_id, cache_base))
        layers.append(targz)

    # Does the user want to change permissions of tar?
    func2 = None
    if PLUGIN_FIXPERMS:
        func2 = change_permissions

    if len(tasks) > 0:
        download_layers = download_client.run(func=download_layer,
                                              func2=func2,
                                              tasks=tasks)

    # Get Docker runscript
    runscript = extract_runscript(manifest=client.manifestv1,
                                  includecmd=INCLUDE_CMD)

    # Add the environment export
    tar_file = extract_metadata_tar(client.manifestv1,
                                    client.assemble_uri(),
                                    runscript=runscript)

    bot.verbose2('Tar file with Docker env and labels: %s' % tar_file)

    # Write all layers to the layerfile
    if layerfile is not None:
        bot.verbose3("Writing Docker layers files to %s" % layerfile)
        write_file(layerfile, "\n".join(layers), mode="w")
        if tar_file is not None:
            write_file(layerfile, "\n%s" % tar_file, mode="a")

    # Return additions dictionary
    additions = {"layers": layers,
                 "image": image,
                 "manifest": client.manifest,
                 "manifestv1": client.manifestv1,
                 "cache_base": cache_base,
                 "metadata": tar_file}

    bot.debug("*** FINISHING DOCKER IMPORT PYTHON PORTION ****\n")

    return additions



'''

tasks.py: Tasks for the Docker API

Copyright (c) 2017, Vanessa Sochat. All rights reserved.


'''

import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),  # noqa
                os.path.pardir)))  # noqa
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) # noqa

from sutils import (
    add_http,
    get_cache,
    create_tar,
    change_tar_permissions,
    print_json,
    write_singularity_infos
)

from defaults import (
    DOCKER_NUMBER,
    DOCKER_PREFIX,
    ENV_BASE,
    LABELFILE,
    PLUGIN_FIXPERMS,
    METADATA_FOLDER_NAME,
    RUNSCRIPT_COMMAND_ASIS
)

from helpers.json.main import ADD
from message import bot
from templates import get_template

import json
import re


def download_layer(client, image_id, cache_base):
    '''download_layer is a function (external to the client)
    to download a layer and update the client's token after.
    This is intended to be used by the multiprocessing function.
    If return_tmp is True, the temporary file is returned
    (intended to be renamed later)'''
    targz = client.get_layer(image_id=image_id,
                             download_folder=cache_base,  # return tmp when
                             return_tmp=PLUGIN_FIXPERMS)  # fix permissions

    client.update_token()
    return targz


def change_permissions(tar_file, file_permission=None, folder_permission=None):
    '''change_permissions is a wrapper for change_tar_permissions,
    intended for use as a function for multiprocessing. To ensure atomic
    download and permission changes, the input file here is expected
    to have a temporary extension. This wrapper simply calls the function
    to change_tar_permissions, and then renames to the final file
    '''
    fixed_tar = change_tar_permissions(tar_file,
                                       file_permission=file_permission,
                                       folder_permission=folder_permission)

    final_tar = "%s.tar.gz" % fixed_tar.split('.tar.gz')[0]
    os.rename(fixed_tar, final_tar)
    return final_tar


def extract_runscript(manifest, includecmd=False):
    '''create_runscript will write a bash script with default "ENTRYPOINT"
    into the base_dir. If includecmd is True, CMD is used instead. For both.
    if the result is found empty, the other is tried, and then a default used.
    :param manifest: the manifest to use to get the runscript
    :param includecmd: overwrite default command (ENTRYPOINT) default is False
    '''
    cmd = None

    # Does the user want to use the CMD instead of ENTRYPOINT?
    commands = ["Entrypoint", "Cmd"]
    if includecmd is True:
        commands.reverse()
    configs = get_configs(manifest, commands)

    # Look for non "None" command
    for command in commands:
        if configs[command] is not None:
            cmd = configs[command]
            break

    if cmd is not None:
        bot.verbose3("Adding Docker %s as Singularity runscript..."
                     % command.upper())

        # If the command is a list, join. (eg ['/usr/bin/python','hello.py']
        if not isinstance(cmd, list):
            cmd = [cmd]

        cmd = " ".join(['"%s"' % x for x in cmd])

        if not RUNSCRIPT_COMMAND_ASIS:
            cmd = 'exec %s "$@"' % cmd
        cmd = "#!/bin/sh\n\n%s\n" % cmd
        return cmd

    bot.debug("CMD and ENTRYPOINT not found, skipping runscript.")
    return cmd


def extract_metadata_tar(manifest,
                         image_name,
                         include_env=True,
                         include_labels=True,
                         runscript=None):

    '''extract_metadata_tar will write a tarfile with the environment,
    labels, and runscript. include_env and include_labels should be booleans,
    and runscript should be None or a string to write to the runscript.
    '''
    tar_file = None
    files = []
    if include_env or include_labels:

        # Extract and add environment
        if include_env:
            environ = extract_env(manifest)
            if environ not in [None, ""]:
                bot.verbose3('Adding Docker environment to metadata tar')
                template = get_template('tarinfo')
                template['name'] = './%s/env/%s-%s.sh' % (METADATA_FOLDER_NAME,
                                                          DOCKER_NUMBER,
                                                          DOCKER_PREFIX)
                template['content'] = environ
                files.append(template)

        # Extract and add labels
        if include_labels:
            labels = extract_labels(manifest)
            if labels is not None:
                if isinstance(labels, dict):
                    labels = print_json(labels)
                bot.verbose3('Adding Docker labels to metadata tar')
                template = get_template('tarinfo')
                template['name'] = "./%s/labels.json" % METADATA_FOLDER_NAME
                template['content'] = labels
                files.append(template)

        if runscript is not None:
            bot.verbose3('Adding Docker runscript to metadata tar')
            template = get_template('tarinfo')
            template['name'] = "./%s/runscript" % METADATA_FOLDER_NAME
            template['content'] = runscript
            files.append(template)

    if len(files) > 0:
        output_folder = get_cache(subfolder="metadata", quiet=True)
        tar_file = create_tar(files, output_folder)
    else:
        bot.warning("No metadata will be included.")
    return tar_file


def extract_env(manifest):
    '''extract the environment from the manifest, or return None.
    Used by functions env_extract_image, and env_extract_tar
    '''
    environ = get_config(manifest, 'Env')
    if environ is not None:
        if not isinstance(environ, list):
            environ = [environ]

        lines = []
        for line in environ:
            line = re.findall("(?P<var_name>.+?)=(?P<var_value>.+)", line)
            line = ['export %s="%s"' % (x[0], x[1]) for x in line]
            lines = lines + line

        environ = "\n".join(lines)
        bot.verbose3("Found Docker container environment!")
    return environ


def env_extract_image(manifest):
    '''env_extract_image will write a file of key value pairs
    of the environment to export. The manner to export must
    be determined by the calling process depending on the OS type.
    :param manifest: the manifest to use
    '''
    environ = extract_env(manifest)
    if environ is not None:
        environ_file = write_singularity_infos(base_dir=ENV_BASE,
                                               prefix=DOCKER_PREFIX,
                                               start_number=DOCKER_NUMBER,
                                               content=environ,
                                               extension='sh')
    return environ


def extract_labels(manifest, labelfile=None, prefix=None):
    '''extract_labels will write a file of key value pairs including
    maintainer, and labels.
    :param manifest: the manifest to use
    :param labelfile: if defined, write to labelfile (json)
    :param prefix: an optional prefix to add to the names
    '''
    if prefix is None:
        prefix = ""

    labels = get_config(manifest, 'Labels')
    if labels is not None and len(labels) is not 0:
        bot.verbose3("Found Docker container labels!")
        if labelfile is not None:
            for key, value in labels.items():
                key = "%s%s" % (prefix, key)
                value = ADD(key, value, labelfile, force=True)
    return labels


def get_config(manifest, spec="Entrypoint", delim=None):
    '''get_config returns a particular spec (default is Entrypoint)
    from a VERSION 1 manifest obtained with get_manifest.
    :param manifest: the manifest obtained from get_manifest
    :param spec: the key of the spec to return, default is "Entrypoint"
    :param delim: Given a list, the delim to use to join the entries.
                  Default is newline
    '''
    cmd = None
    if "history" in manifest:
        history = manifest['history']
        for entry in manifest['history']:
            if 'v1Compatibility' in entry:
                entry = json.loads(entry['v1Compatibility'])
                if "config" in entry:
                    if spec in entry["config"]:
                        if entry["config"][spec] is not None:
                            cmd = entry["config"][spec]
                            break

    # Standard is to include commands like ['/bin/sh']
    if isinstance(cmd, list):
        if delim is not None:
            cmd = delim.join(cmd)
    bot.verbose3("Found Docker command (%s) %s" % (spec, cmd))

    return cmd


def get_configs(manifest, keys, delim=None):
    '''get_configs is a wrapper for get_config to return a dictionary
    with multiple config items.
    :param manifest: the complete manifest
    :param keys: the key to find
    :param delim: given a list, combine based on this delim
    '''
    configs = dict()
    if not isinstance(keys, list):
        keys = [keys]
    for key in keys:
        configs[key] = get_config(manifest, key, delim=delim)
    return configs



# -*- coding: utf-8 -*-

'''

message.py: simple logger for Singularity python helper

The error names (prefix) and integer level are assigned as follows:

ABRT -4
ERROR -3
WARNING -2
LOG -1
INFO 1
QUIET 0
VERBOSE 2
VERBOSE1 2
VERBOSE2 3
VERBOSE3 4
DEBUG 5

VERBOSE is equivalent to VERBOSE1 (this is mirroring the C code)
and for each level, calling it corresponds to calling the class'
function for it. E.g., DEBUG --> bot.debug('This is the message!')

The following levels are to stderr:

5,4,3,2,1,-1,-2,-3,-4

The following levels are only to stdout

1

The following levels do nothing (quiet)

0

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import os
import sys

ABRT = -4
ERROR = -3
WARNING = -2
LOG = -1
INFO = 1
QUIET = 0
VERBOSE = VERBOSE1 = 2
VERBOSE2 = 3
VERBOSE3 = 4
DEBUG = 5


class SingularityMessage:

    def __init__(self, MESSAGELEVEL=None):
        self.level = get_logging_level()
        self.history = []
        self.errorStream = sys.stderr
        self.outputStream = sys.stdout
        self.colorize = self.useColor()
        self.colors = {ABRT: "\033[31m",     # dark red
                       ERROR: "\033[91m",    # red
                       WARNING: "\033[93m",  # dark yellow
                       LOG: "\033[95m",      # purple
                       DEBUG: "\033[36m",    # cyan
                       'OFF': "\033[0m"}     # end sequence

    # Colors --------------------------------------------

    def useColor(self):
        '''useColor will determine if color should be added
        to a print. Will check if being run in a terminal, and
        if has support for asci'''
        COLORIZE = get_user_color_preference()
        if COLORIZE is not None:
            return COLORIZE
        streams = [self.errorStream, self.outputStream]
        for stream in streams:
            if not hasattr(stream, 'isatty'):
                return False
            if not stream.isatty():
                return False
        return True

    def addColor(self, level, text):
        '''addColor to the prompt (usually prefix)
        if terminal supports, and specified to do so'''
        if self.colorize:
            if level in self.colors:
                text = "%s%s%s" % (self.colors[level],
                                   text,
                                   self.colors["OFF"])
        return text

    def emitError(self, level):
        '''determine if a level should print to
        stderr, includes all levels but INFO and QUIET'''
        if level in [ABRT,
                     ERROR,
                     WARNING,
                     LOG,
                     VERBOSE,
                     VERBOSE1,
                     VERBOSE2,
                     VERBOSE3,
                     DEBUG]:
            return True
        return False

    def emitOutput(self, level):
        '''determine if a level should print to stdout
        only includes INFO'''
        if level in [INFO]:
            return True
        return False

    def isEnabledFor(self, messageLevel):
        '''check if a messageLevel is enabled to emit a level
        '''
        if messageLevel <= self.level:
            return True
        return False

    def emit(self, level, message, prefix=None):
        '''emit is the main function to print the message
        optionally with a prefix
        :param level: the level of the message
        :param message: the message to print
        :param prefix: a prefix for the message
        '''

        if prefix is not None:
            prefix = self.addColor(level, "%s " % (prefix))
        else:
            prefix = ""
            message = self.addColor(level, message)

        # Add the prefix
        message = "%s%s" % (prefix, message)

        if not message.endswith('\n'):
            message = "%s\n" % (message)

        # If the level is quiet, only print to error
        if self.level == QUIET:
            pass

        # Otherwise if in range print to stdout and stderr
        elif self.isEnabledFor(level):
            if self.emitError(level):
                self.write(self.errorStream, message)
            else:
                self.write(self.outputStream, message)

        # Add all log messages to history
        self.history.append(message)

    def write(self, stream, message):
        '''write will write a message to a stream,
        first checking the encoding
        '''
        if isinstance(message, bytes):
            message = message.decode('utf-8')
        stream.write(message)

    def get_logs(self, join_newline=True):
        ''''get_logs will return the complete history,
        joined by newline (default) or as is.
        '''
        if join_newline:
            return '\n'.join(self.history)
        return self.history

    def show_progress(self,
                      iteration,
                      total,
                      length=40,
                      min_level=0,
                      prefix=None,
                      carriage_return=True,
                      suffix=None,
                      symbol=None):

        '''create a terminal progress bar,
        default bar shows for verbose+
        :param iteration: current iteration (Int)
        :param total: total iterations (Int)
        :param length: character length of bar (Int)
        '''
        perc = 100 * (iteration / float(total))
        progress = int(length * iteration // total)

        if suffix is None:
            suffix = ''

        if prefix is None:
            prefix = 'Progress'

        # Download sizes can be imperfect, setting carriage_return to False
        # and writing newline with caller cleans up the UI
        if perc >= 100:
            perc = 100
            progress = length

        if symbol is None:
            symbol = "="

        if progress < length:
            bar = symbol * progress + '|' + '-' * (length - progress - 1)
        else:
            bar = symbol * progress + '-' * (length - progress)

        # Only show progress bar for level > min_level
        if self.level > min_level:
            perc = "%5s" % ("{0:.1f}").format(perc)
            output = '\r' + prefix + " |%s| %s%s %s" % (bar, perc, '%', suffix)
            sys.stdout.write(output),
            if iteration == total and carriage_return:
                sys.stdout.write('\n')
            sys.stdout.flush()

    def abort(self, message):
        self.emit(ABRT, message, 'ABRT')

    def error(self, message):
        self.emit(ERROR, message, 'ERROR')

    def warning(self, message):
        self.emit(WARNING, message, 'WARNING')

    def log(self, message):
        self.emit(LOG, message, 'LOG')

    def info(self, message):
        self.emit(INFO, message)

    def verbose(self, message):
        self.emit(VERBOSE, message, "VERBOSE")

    def verbose1(self, message):
        self.emit(VERBOSE, message, "VERBOSE1")

    def verbose2(self, message):
        self.emit(VERBOSE2, message, 'VERBOSE2')

    def verbose3(self, message):
        self.emit(VERBOSE3, message, 'VERBOSE3')

    def debug(self, message):
        self.emit(DEBUG, message, 'DEBUG')

    def is_quiet(self):
        '''is_quiet returns true if the level is under 1
        '''
        if self.level < 1:
            return False
        return True


def get_logging_level():
    '''get_logging_level will configure a logging to
    standard out based on the user's selected level,
    which should be in an environment variable called
    SINGULARITY_MESSAGELEVEL. if SINGULARITY_MESSAGELEVEL
    is not set, the maximum level (5) is assumed (all
    messages).

    #define ABRT -4
    #define ERROR -3
    #define WARNING -2
    #define LOG -1
    #define INFO 1

    implied define: QUIET 0

    #define VERBOSE 2
    #define VERBOSE1 2
    #define VERBOSE2 3
    #define VERBOSE3 4
    #define DEBUG 5
    '''

    return int(os.environ.get("SINGULARITY_MESSAGELEVEL", 5))


def get_user_color_preference():
    COLORIZE = os.environ.get('SINGULARITY_COLORIZE', None)
    if COLORIZE is not None:
        COLORIZE = convert2boolean(COLORIZE)
    return COLORIZE


def convert2boolean(arg):
    '''convert2boolean is used for environmental variables
    that must be returned as boolean'''
    if not isinstance(arg, bool):
        return arg.lower() in ("yes", "true", "t", "1", "y")
    return arg


bot = SingularityMessage()



'''

defaults.py: this script acts as a gateway between variables defined at
runtime, and defaults. Any variable that has an unchanging default value
can be found here. The order of operations works as follows:

    1. First preference goes to environment variable set at runtime
    2. Second preference goes to default defined in this file
    3. Then, if neither is found, null is returned except in the
       case that required = True. A required = True variable not found
       will system exit with an error.

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

from message import bot
import tempfile
import os
import pwd
import sys


def getenv(variable_key, default=None, required=False, silent=False):
    '''getenv will attempt to get an environment variable. If the
    variable is not found, None is returned.
    :param variable_key: the variable name
    :param required: exit with error if not found
    :param silent: Do not print debugging information for variable
    '''
    variable = os.environ.get(variable_key, default)
    if variable is None and required:
        bot.error("Cannot find environment variable %s, exiting."
                  % (variable_key))
        sys.exit(1)

    if silent:
        if variable is not None:
            bot.verbose2("%s found" % (variable_key))
    else:
        if variable is not None:
            bot.verbose2("%s found as %s" % (variable_key, variable))
        else:
            bot.verbose2("%s not defined (None)" % (variable_key))

    return variable


def convert2boolean(arg):
    '''convert2boolean is used for environmental variables
    that must be returned as boolean'''
    if not isinstance(arg, bool):
        return arg.lower() in ("yes", "true", "t", "1", "y")
    return arg


#######################################################################
# Singularity
#######################################################################

# Filled in to exec %s "$@"
RUNSCRIPT_COMMAND_ASIS = convert2boolean(getenv("SINGULARITY_COMMAND_ASIS",
                                         default=False))

SINGULARITY_ROOTFS = getenv("SINGULARITY_ROOTFS")
METADATA_FOLDER_NAME = ".singularity.d"
_metadata_base = "%s/%s" % (SINGULARITY_ROOTFS, METADATA_FOLDER_NAME)
METADATA_BASE = getenv("SINGULARITY_METADATA_FOLDER", _metadata_base,
                       required=True)


#######################################################################
# Plugins and Formatting
#######################################################################

PLUGIN_FIXPERMS = convert2boolean(getenv("SINGULARITY_FIX_PERMS", False))

COLORIZE = getenv("SINGULARITY_COLORIZE", None)
if COLORIZE is not None:
    COLORIZE = convert2boolean(COLORIZE)

#######################################################################
# Cache
#######################################################################

DISABLE_CACHE = convert2boolean(getenv("SINGULARITY_DISABLE_CACHE",
                                default=False))

if DISABLE_CACHE is True:
    SINGULARITY_CACHE = tempfile.mkdtemp()
else:
    userhome = pwd.getpwuid(os.getuid())[5]
    _cache = os.path.join(userhome, ".singularity")
    SINGULARITY_CACHE = getenv("SINGULARITY_CACHEDIR", default=_cache)


#######################################################################
# Docker
#######################################################################

# API
DOCKER_API_BASE = "index.docker.io"  # registry
CUSTOM_REGISTRY = getenv("REGISTRY")
NAMESPACE = "library"
CUSTOM_NAMESPACE = getenv('NAMESPACE')
DOCKER_API_VERSION = "v2"
DOCKER_ARCHITECTURE = getenv("SINGULARITY_DOCKER_ARCHITECTURE", "amd64")
DOCKER_OS = getenv("SINGULARITY_DOCKER_OS", "linux")
TAG = "latest"

# Container Metadata
DOCKER_NUMBER = 10  # number to start docker files at in ENV_DIR
DOCKER_PREFIX = "docker"
SHUB_PREFIX = "shub"

# Defaults for environment, runscript, labels
_envbase = "%s/env" % (METADATA_BASE)
_runscript = "%s/singularity" % (SINGULARITY_ROOTFS)
_environment = "%s/90-environment.sh" % (_envbase)
_labelfile = "%s/labels.json" % (METADATA_BASE)
_helpfile = "%s/runscript.help" % (METADATA_BASE)
_deffile = "%s/Singularity" % (METADATA_BASE)
_testfile = "%s/test" % (METADATA_BASE)


ENVIRONMENT = getenv("SINGULARITY_ENVIRONMENT", _environment)
RUNSCRIPT = getenv("SINGULARITY_RUNSCRIPT", _runscript)
TESTFILE = getenv("SINGULARITY_TESTFILE", _testfile)
DEFFILE = getenv("SINGULARITY_DEFFILE", _deffile)
HELPFILE = getenv("SINGULARITY_HELPFILE", _helpfile)
ENV_BASE = getenv("SINGULARITY_ENVBASE", _envbase)
LABELFILE = getenv("SINGULARITY_LABELFILE", _labelfile)
INCLUDE_CMD = convert2boolean(getenv("SINGULARITY_INCLUDECMD", False))
DISABLE_HTTPS = convert2boolean(getenv("SINGULARITY_NOHTTPS", False))

#######################################################################
# Singularity Hub
#######################################################################

SINGULARITY_PULLFOLDER = getenv("SINGULARITY_PULLFOLDER", os.getcwd())
SHUB_API_BASE = "www.singularity-hub.org"
SHUB_NAMEBYHASH = getenv("SHUB_NAMEBYHASH")
SHUB_NAMEBYCOMMIT = getenv("SHUB_NAMEBYCOMMIT")
SHUB_CONTAINERNAME = getenv("SHUB_CONTAINERNAME")

#######################################################################
# Python Internal API URI Handling
#######################################################################

_layerfile = "%s/.layers" % (METADATA_BASE)
LAYERFILE = getenv("SINGULARITY_CONTENTS", _layerfile)

SINGULARITY_WORKERS = int(getenv("SINGULARITY_PYTHREADS", 9))



'''

test_docker_tasks.py: Docker tasks testing for Singularity in Python

Copyright (c) 2017, Vanessa Sochat. All rights reserved.

'''

import os
import re
import sys
sys.path.append('..')  # noqa

from unittest import TestCase
import shutil
import tempfile

VERSION = sys.version_info[0]

print("*** PYTHON VERSION %s TASKS TESTING START ***" % VERSION)


class TestApi(TestCase):

    def setUp(self):
        self.image = 'docker://ubuntu:latest'
        self.tmpdir = tempfile.mkdtemp()
        os.environ['SINGULARITY_ROOTFS'] = self.tmpdir
        os.mkdir('%s/.singularity.d' % self.tmpdir)
        from docker.api import DockerApiConnection
        self.client = DockerApiConnection(image=self.image)

        print("\n---START----------------------------------------")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

        print("---END------------------------------------------")

    def test_create_runscript(self):
        '''test_create_runscript should ensure that
        a runscript is generated with some command
        '''
        from docker.api import DockerApiConnection

        print('Testing creation of runscript')
        from docker.tasks import extract_runscript

        manifest = self.client.get_manifest(old_version=True)

        print("Case 1: Asking for CMD when none defined")
        default_cmd = 'exec "/bin/bash"'
        runscript = extract_runscript(manifest=manifest,
                                      includecmd=True)
        self.assertTrue(default_cmd in runscript)

        print("Case 2: Asking for ENTRYPOINT when none defined")
        runscript = extract_runscript(manifest=manifest)
        self.assertTrue(default_cmd in runscript)

        client = DockerApiConnection(image="docker://bids/mriqc:0.0.2")
        manifest = client.get_manifest(old_version=True)

        print("Case 3: Asking for ENTRYPOINT when defined")
        runscript = extract_runscript(manifest=manifest)
        self.assertTrue('exec "/run_mriqc"' in runscript)

        print("Case 4: Asking for CMD when defined")
        runscript = extract_runscript(manifest=manifest,
                                      includecmd=True)
        self.assertTrue('exec "--help"' in runscript)

        print("Case 5: Asking for ENTRYPOINT when None, should return CMD")
        from docker.tasks import get_configs
        client = DockerApiConnection(image="tensorflow/tensorflow:1.0.0")
        manifest = client.get_manifest(old_version=True)

        configs = get_configs(manifest, ['Cmd', 'Entrypoint'])
        self.assertEqual(configs['Entrypoint'], None)
        runscript = extract_runscript(manifest=manifest)
        self.assertTrue(configs['Cmd'][0] in runscript)

    def test_get_config(self):
        '''test_get_config will obtain parameters
        from the DOcker configuration json
        '''
        from docker.api import DockerApiConnection
        from docker.tasks import get_config

        # Default should return entrypoint
        print("Case 1: Ask for default command (Entrypoint)")
        manifest = self.client.get_manifest(old_version=True)
        entrypoint = get_config(manifest=manifest)

        # Ubuntu latest should have None
        self.assertEqual(entrypoint, None)

        print("Case 2: Ask for custom command (Cmd)")
        entrypoint = get_config(manifest=manifest,
                                spec="Cmd")
        self.assertTrue('/bin/bash' in entrypoint)


if __name__ == '__main__':
    unittest.main()



'''

test_docker_import_user.py: Docker testing functions for Singularity in Python

Copyright (c) 2016-2017, Vanessa Sochat. All rights reserved.

"Singularity" Copyright (c) 2016, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

This software is licensed under a customized 3-clause BSD license.  Please
consult LICENSE file distributed with the sources of this project regarding
your rights to use or distribute this software.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights. As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.

'''

import os
import sys
sys.path.append('..')  # noqa

from subprocess import (
    Popen,
    PIPE,
    STDOUT
)

from unittest import TestCase
import shutil
import tempfile


VERSION = sys.version_info[0]

print("*** PYTHON VERSION %s DOCKER IMPORT TESTING START ***" % VERSION)


class TestImport(TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.contents_file = "%s/hello-kitty" % self.tmpdir
        self.here = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                 os.path.pardir))

        # Variables are obtained from environment
        os.environ["SINGULARITY_CONTAINER"] = "docker://ubuntu:latest"
        os.environ["SINGULARITY_CONTENTS"] = self.contents_file

        print("\n---START----------------------------------------")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        print("---END------------------------------------------")

    def test_IMPORT(self):
        '''test_import will test the IMPORT USER function
        '''
        script_path = "%s/import.py" % self.here
        if VERSION == 2:
            testing_command = ["python2", script_path]
        else:
            testing_command = ["python3", script_path]

        output = Popen(testing_command, stderr=STDOUT, stdout=PIPE)
        t = output.communicate()[0], output.returncode
        result = {'message': t[0],
                  'return_code': t[1]}
        self.assertEqual(result['return_code'], 0)
        self.assertTrue(os.path.exists(self.contents_file))


if __name__ == '__main__':
    unittest.main()
